@@ -34,7 +34,7 @@ using IKVM.Attributes;
 
 namespace IKVM.Internal
 {
-	class AotTypeWrapper : DynamicTypeWrapper
+	sealed class AotTypeWrapper : DynamicTypeWrapper
 	{
 		private FieldInfo ghostRefField;
 		private MethodBuilder ghostIsInstanceMethod;
@@ -43,6 +43,7 @@ namespace IKVM.Internal
 		private MethodBuilder ghostCastArrayMethod;
 		private TypeBuilder typeBuilderGhostInterface;
 		private Annotation annotation;
+		private Type enumType;
 		private MethodWrapper[] replacedMethods;
 		private WorkaroundBaseClass workaroundBaseClass;
 
@@ -82,7 +83,8 @@ namespace IKVM.Internal
 					{
 						name = "_" + name;
 					}
-					TypeBuilder typeBuilder = classLoader.GetTypeWrapperFactory().ModuleBuilder.DefineType(name, TypeAttributes.Public | TypeAttributes.Abstract, base.GetBaseTypeForDefineType());
+					TypeWrapperFactory context = classLoader.GetTypeWrapperFactory();
+					TypeBuilder typeBuilder = context.ModuleBuilder.DefineType(name, TypeAttributes.Public | TypeAttributes.Abstract, base.GetBaseTypeForDefineType());
 					AttributeHelper.HideFromJava(typeBuilder);
 					AttributeHelper.SetEditorBrowsableNever(typeBuilder);
 					workaroundBaseClass = new WorkaroundBaseClass(this, typeBuilder, methods.ToArray());
@@ -91,7 +93,7 @@ namespace IKVM.Internal
 					{
 						if (ReferenceEquals(mw.Name, StringConstants.INIT) && mw.IsAccessibleFrom(baseTypeWrapper, this, this))
 						{
-							constructors.Add(new ConstructorForwarder(typeBuilder, mw));
+							constructors.Add(new ConstructorForwarder(context, typeBuilder, mw));
 						}
 					}
 					replacedMethods = constructors.ToArray();
@@ -139,7 +141,7 @@ namespace IKVM.Internal
 				{
 					foreach (MethodWrapper mw in methods)
 					{
-						MethodBuilder mb = typeBuilder.DefineMethod(mw.Name, MethodAttributes.FamORAssem | MethodAttributes.Virtual | MethodAttributes.HideBySig | MethodAttributes.CheckAccessOnOverride, mw.ReturnTypeForDefineMethod, mw.GetParametersForDefineMethod());
+						MethodBuilder mb = mw.GetDefineMethodHelper().DefineMethod(wrapper, typeBuilder, mw.Name, MethodAttributes.FamORAssem | MethodAttributes.Virtual | MethodAttributes.HideBySig | MethodAttributes.CheckAccessOnOverride);
 						AttributeHelper.HideFromJava(mb);
 						CodeEmitter ilgen = CodeEmitter.Create(mb);
 						ilgen.EmitThrow("java.lang.AbstractMethodError");
@@ -152,13 +154,15 @@ namespace IKVM.Internal
 
 		private sealed class ConstructorForwarder : MethodWrapper
 		{
+			private readonly TypeWrapperFactory context;
 			private readonly TypeBuilder typeBuilder;
 			private readonly MethodWrapper ctor;
-			private ConstructorBuilder constructorBuilder;
+			private MethodBuilder constructorBuilder;
 
-			internal ConstructorForwarder(TypeBuilder typeBuilder, MethodWrapper ctor)
+			internal ConstructorForwarder(TypeWrapperFactory context, TypeBuilder typeBuilder, MethodWrapper ctor)
 				: base(ctor.DeclaringType, ctor.Name, ctor.Signature, null, null, null, ctor.Modifiers, MemberFlags.None)
 			{
+				this.context = context;
 				this.typeBuilder = typeBuilder;
 				this.ctor = ctor;
 			}
@@ -166,14 +170,14 @@ namespace IKVM.Internal
 			protected override void DoLinkMethod()
 			{
 				ctor.Link();
-				Type[] parameters = ctor.GetParametersForDefineMethod();
-				constructorBuilder = typeBuilder.DefineConstructor(MethodAttributes.PrivateScope, CallingConventions.Standard, parameters);
+				DefineMethodHelper dmh = ctor.GetDefineMethodHelper();
+				constructorBuilder = dmh.DefineConstructor(context, typeBuilder, MethodAttributes.PrivateScope);
 				AttributeHelper.HideFromJava(constructorBuilder);
 				CodeEmitter ilgen = CodeEmitter.Create(constructorBuilder);
 				ilgen.Emit(OpCodes.Ldarg_0);
-				for (int i = 1; i <= parameters.Length; i++)
+				for (int i = 1; i <= dmh.ParameterCount; i++)
 				{
-					ilgen.Emit(OpCodes.Ldarg_S, (byte)i);
+					ilgen.EmitLdarg(i);
 				}
 				ctor.EmitCall(ilgen);
 				ilgen.Emit(OpCodes.Ret);
@@ -438,14 +442,14 @@ namespace IKVM.Internal
 						MethodBuilder mb = mw.GetMethod() as MethodBuilder;
 						if(mb == null || mb.DeclaringType != typeBuilder || (!mb.IsFinal && final))
 						{
-							mb = typeBuilder.DefineMethod(GenerateUniqueMethodName("get_" + prop.Name, mw), GetPropertyMethodAttributes(mw, final), typeWrapper.TypeAsSignatureType, indexer);
+							mb = typeBuilder.DefineMethod("get_" + prop.Name, GetPropertyMethodAttributes(mw, final), typeWrapper.TypeAsSignatureType, indexer);
 							AttributeHelper.HideFromJava(mb);
 							CodeEmitter ilgen = CodeEmitter.Create(mb);
 							if(mw.IsStatic)
 							{
 								for(int i = 0; i < indexer.Length; i++)
 								{
-									ilgen.Emit(OpCodes.Ldarg, (short)i);
+									ilgen.EmitLdarg(i);
 								}
 								mw.EmitCall(ilgen);
 							}
@@ -454,7 +458,7 @@ namespace IKVM.Internal
 								ilgen.Emit(OpCodes.Ldarg_0);
 								for(int i = 0; i < indexer.Length; i++)
 								{
-									ilgen.Emit(OpCodes.Ldarg, (short)(i + 1));
+									ilgen.EmitLdarg(i + 1);
 								}
 								mw.EmitCallvirt(ilgen);
 							}
@@ -479,14 +483,14 @@ namespace IKVM.Internal
 						MethodBuilder mb = mw.GetMethod() as MethodBuilder;
 						if(mb == null || mb.DeclaringType != typeBuilder || (!mb.IsFinal && final))
 						{
-							mb = typeBuilder.DefineMethod(GenerateUniqueMethodName("set_" + prop.Name, mw), GetPropertyMethodAttributes(mw, final), mw.ReturnTypeForDefineMethod, args);
+							mb = typeBuilder.DefineMethod("set_" + prop.Name, GetPropertyMethodAttributes(mw, final), mw.ReturnTypeForDefineMethod, args);
 							AttributeHelper.HideFromJava(mb);
 							CodeEmitter ilgen = CodeEmitter.Create(mb);
 							if(mw.IsStatic)
 							{
 								for(int i = 0; i <= indexer.Length; i++)
 								{
-									ilgen.Emit(OpCodes.Ldarg, (short)i);
+									ilgen.EmitLdarg(i);
 								}
 								mw.EmitCall(ilgen);
 							}
@@ -495,7 +499,7 @@ namespace IKVM.Internal
 								ilgen.Emit(OpCodes.Ldarg_0);
 								for(int i = 0; i <= indexer.Length; i++)
 								{
-									ilgen.Emit(OpCodes.Ldarg, (short)(i + 1));
+									ilgen.EmitLdarg(i + 1);
 								}
 								mw.EmitCallvirt(ilgen);
 							}
@@ -662,7 +666,7 @@ namespace IKVM.Internal
 								Type returnType;
 								Type[] parameterTypes;
 								MapSignature(constructor.Sig, out returnType, out parameterTypes);
-								ConstructorBuilder cb = typeBuilder.DefineConstructor(attribs, CallingConventions.Standard, parameterTypes);
+								MethodBuilder cb = ReflectUtil.DefineConstructor(typeBuilder, attribs, parameterTypes);
 								if(setmodifiers)
 								{
 									AttributeHelper.SetModifiers(cb, (Modifiers)constructor.Modifiers, false);
@@ -688,7 +692,7 @@ namespace IKVM.Internal
 								{
 									if(mw.Name == "<init>" && mw.Signature == constructor.Sig)
 									{
-										ConstructorBuilder mb = mw.GetMethod() as ConstructorBuilder;
+										MethodBuilder mb = mw.GetMethod() as MethodBuilder;
 										if(mb != null)
 										{
 											foreach(IKVM.Internal.MapXml.Attribute attr in constructor.Attributes)
@@ -783,7 +787,7 @@ namespace IKVM.Internal
 										throw new InvalidOperationException("Method " + m.Name + m.Sig + " not found in interface " + tw.Name);
 									}
 									mw.Link();
-									MethodBuilder mb = typeBuilder.DefineMethod(tw.Name + "/" + m.Name, MethodAttributes.Private | MethodAttributes.NewSlot | MethodAttributes.Virtual | MethodAttributes.Final | MethodAttributes.CheckAccessOnOverride, mw.ReturnTypeForDefineMethod, mw.GetParametersForDefineMethod());
+									MethodBuilder mb = mw.GetDefineMethodHelper().DefineMethod(this, typeBuilder, tw.Name + "/" + m.Name, MethodAttributes.Private | MethodAttributes.NewSlot | MethodAttributes.Virtual | MethodAttributes.Final | MethodAttributes.CheckAccessOnOverride);
 									AttributeHelper.HideFromJava(mb);
 									typeBuilder.DefineMethodOverride(mb, (MethodInfo)mw.GetMethod());
 									CodeEmitter ilgen = CodeEmitter.Create(mb);
@@ -801,7 +805,7 @@ namespace IKVM.Internal
 		{
 			if(typeBuilderGhostInterface != null)
 			{
-				return typeBuilderGhostInterface.DefineMethod(name, attribs, mw.ReturnTypeForDefineMethod, mw.GetParametersForDefineMethod());
+				return mw.GetDefineMethodHelper().DefineMethod(this, typeBuilderGhostInterface, name, attribs);
 			}
 			return null;
 		}
@@ -817,7 +821,7 @@ namespace IKVM.Internal
 					if(!methods[i].IsStatic)
 					{
 						TypeWrapper[] args = methods[i].GetParameters();
-						MethodBuilder stub = typeBuilder.DefineMethod(methods[i].Name, MethodAttributes.Public, methods[i].ReturnTypeForDefineMethod, methods[i].GetParametersForDefineMethod());
+						MethodBuilder stub = methods[i].GetDefineMethodHelper().DefineMethod(this, typeBuilder, methods[i].Name, MethodAttributes.Public);
 						AddParameterMetadata(stub, methods[i]);
 						AttributeHelper.SetModifiers(stub, methods[i].Modifiers, methods[i].IsInternal);
 						CodeEmitter ilgen = CodeEmitter.Create(stub);
@@ -828,29 +832,29 @@ namespace IKVM.Internal
 						ilgen.Emit(OpCodes.Dup);
 						ilgen.Emit(OpCodes.Isinst, typeBuilderGhostInterface);
 						CodeEmitterLabel label = ilgen.DefineLabel();
-						ilgen.Emit(OpCodes.Brfalse_S, label);
+						ilgen.EmitBrfalse(label);
 						ilgen.Emit(OpCodes.Castclass, typeBuilderGhostInterface);
 						for(int k = 0; k < args.Length; k++)
 						{
-							ilgen.Emit(OpCodes.Ldarg_S, (byte)(k + 1));
+							ilgen.EmitLdarg(k + 1);
 						}
 						ilgen.Emit(OpCodes.Callvirt, (MethodInfo)methods[i].GetMethod());
-						ilgen.Emit(OpCodes.Br, end);
+						ilgen.EmitBr(end);
 						ilgen.MarkLabel(label);
 						for(int j = 0; j < implementers.Length; j++)
 						{
 							ilgen.Emit(OpCodes.Dup);
 							ilgen.Emit(OpCodes.Isinst, implementers[j].TypeAsTBD);
 							label = ilgen.DefineLabel();
-							ilgen.Emit(OpCodes.Brfalse_S, label);
+							ilgen.EmitBrfalse(label);
 							ilgen.Emit(OpCodes.Castclass, implementers[j].TypeAsTBD);
 							for(int k = 0; k < args.Length; k++)
 							{
-								ilgen.Emit(OpCodes.Ldarg_S, (byte)(k + 1));
+								ilgen.EmitLdarg(k + 1);
 							}
 							MethodWrapper mw = implementers[j].GetMethodWrapper(methods[i].Name, methods[i].Signature, true);
 							mw.EmitCallvirt(ilgen);
-							ilgen.Emit(OpCodes.Br, end);
+							ilgen.EmitBr(end);
 							ilgen.MarkLabel(label);
 						}
 						// we need to do a null check (null fails all the isinst checks)
@@ -893,9 +897,9 @@ namespace IKVM.Internal
 						ilgen.Emit(OpCodes.Ldarg_0);
 						ilgen.Emit(OpCodes.Isinst, implementers[i].TypeAsTBD);
 						CodeEmitterLabel label = ilgen.DefineLabel();
-						ilgen.Emit(OpCodes.Brfalse_S, label);
+						ilgen.EmitBrfalse(label);
 						ilgen.Emit(OpCodes.Ldc_I4_1);
-						ilgen.Emit(OpCodes.Br, end);
+						ilgen.EmitBr(end);
 						ilgen.MarkLabel(label);
 					}
 					ilgen.Emit(OpCodes.Ldarg_0);
@@ -913,7 +917,7 @@ namespace IKVM.Internal
 					CodeEmitterLocal localRank = ilgen.DeclareLocal(Types.Int32);
 					ilgen.Emit(OpCodes.Ldarg_0);
 					CodeEmitterLabel skip = ilgen.DefineLabel();
-					ilgen.Emit(OpCodes.Brtrue_S, skip);
+					ilgen.EmitBrtrue(skip);
 					ilgen.Emit(OpCodes.Ldc_I4_0);
 					ilgen.Emit(OpCodes.Ret);
 					ilgen.MarkLabel(skip);
@@ -923,7 +927,7 @@ namespace IKVM.Internal
 					ilgen.Emit(OpCodes.Ldarg_1);
 					ilgen.Emit(OpCodes.Stloc, localRank);
 					skip = ilgen.DefineLabel();
-					ilgen.Emit(OpCodes.Br_S, skip);
+					ilgen.EmitBr(skip);
 					CodeEmitterLabel iter = ilgen.DefineLabel();
 					ilgen.MarkLabel(iter);
 					ilgen.Emit(OpCodes.Ldloc, localType);
@@ -935,11 +939,11 @@ namespace IKVM.Internal
 					ilgen.Emit(OpCodes.Stloc, localRank);
 					ilgen.Emit(OpCodes.Ldloc, localRank);
 					CodeEmitterLabel typecheck = ilgen.DefineLabel();
-					ilgen.Emit(OpCodes.Brfalse_S, typecheck);
+					ilgen.EmitBrfalse(typecheck);
 					ilgen.MarkLabel(skip);
 					ilgen.Emit(OpCodes.Ldloc, localType);
 					ilgen.Emit(OpCodes.Callvirt, Types.Type.GetMethod("get_IsArray"));
-					ilgen.Emit(OpCodes.Brtrue_S, iter);
+					ilgen.EmitBrtrue(iter);
 					ilgen.Emit(OpCodes.Ldc_I4_0);
 					ilgen.Emit(OpCodes.Ret);
 					ilgen.MarkLabel(typecheck);
@@ -950,7 +954,7 @@ namespace IKVM.Internal
 						ilgen.Emit(OpCodes.Ldloc, localType);
 						ilgen.Emit(OpCodes.Callvirt, Types.Type.GetMethod("IsAssignableFrom"));
 						CodeEmitterLabel label = ilgen.DefineLabel();
-						ilgen.Emit(OpCodes.Brfalse_S, label);
+						ilgen.EmitBrfalse(label);
 						ilgen.Emit(OpCodes.Ldc_I4_1);
 						ilgen.Emit(OpCodes.Ret);
 						ilgen.MarkLabel(label);
@@ -960,7 +964,7 @@ namespace IKVM.Internal
 					ilgen.Emit(OpCodes.Ldloc, localType);
 					ilgen.Emit(OpCodes.Callvirt, Types.Type.GetMethod("IsAssignableFrom"));
 					skip = ilgen.DefineLabel();
-					ilgen.Emit(OpCodes.Brfalse_S, skip);
+					ilgen.EmitBrfalse(skip);
 					ilgen.Emit(OpCodes.Ldc_I4_1);
 					ilgen.Emit(OpCodes.Ret);
 					ilgen.MarkLabel(skip);
@@ -980,7 +984,7 @@ namespace IKVM.Internal
 					{
 						ilgen.Emit(OpCodes.Ldarg_0);
 						ilgen.Emit(OpCodes.Isinst, implementers[i].TypeAsTBD);
-						ilgen.Emit(OpCodes.Brtrue, end);
+						ilgen.EmitBrtrue(end);
 					}
 					ilgen.Emit(OpCodes.Ldarg_0);
 					ilgen.Emit(OpCodes.Castclass, typeBuilderGhostInterface);
@@ -1012,11 +1016,11 @@ namespace IKVM.Internal
 					ilgen = CodeEmitter.Create(mb);
 					end = ilgen.DefineLabel();
 					ilgen.Emit(OpCodes.Ldarg_0);
-					ilgen.Emit(OpCodes.Brfalse_S, end);
+					ilgen.EmitBrfalse(end);
 					ilgen.Emit(OpCodes.Ldarg_0);
 					ilgen.Emit(OpCodes.Ldarg_1);
 					ilgen.Emit(OpCodes.Call, ghostIsInstanceArrayMethod);
-					ilgen.Emit(OpCodes.Brtrue_S, end);
+					ilgen.EmitBrtrue(end);
 					ilgen.Emit(OpCodes.Ldarg_0);
 					ilgen.Emit(OpCodes.Ldtoken, typeBuilder);
 					ilgen.Emit(OpCodes.Ldarg_1);
@@ -1050,9 +1054,9 @@ namespace IKVM.Internal
 					mb = typeBuilder.DefineMethod("op_Equality", MethodAttributes.HideBySig | MethodAttributes.Public | MethodAttributes.Static | MethodAttributes.SpecialName, Types.Boolean, new Type[] { typeBuilder, typeBuilder });
 					AttributeHelper.HideFromJava(mb);
 					ilgen = CodeEmitter.Create(mb);
-					ilgen.Emit(OpCodes.Ldarga_S, (byte)0);
+					ilgen.EmitLdarga(0);
 					ilgen.Emit(OpCodes.Ldfld, ghostRefField);
-					ilgen.Emit(OpCodes.Ldarga_S, (byte)1);
+					ilgen.EmitLdarga(1);
 					ilgen.Emit(OpCodes.Ldfld, ghostRefField);
 					ilgen.Emit(OpCodes.Ceq);
 					ilgen.Emit(OpCodes.Ret);
@@ -1062,9 +1066,9 @@ namespace IKVM.Internal
 					mb = typeBuilder.DefineMethod("op_Inequality", MethodAttributes.HideBySig | MethodAttributes.Public | MethodAttributes.Static | MethodAttributes.SpecialName, Types.Boolean, new Type[] { typeBuilder, typeBuilder });
 					AttributeHelper.HideFromJava(mb);
 					ilgen = CodeEmitter.Create(mb);
-					ilgen.Emit(OpCodes.Ldarga_S, (byte)0);
+					ilgen.EmitLdarga(0);
 					ilgen.Emit(OpCodes.Ldfld, ghostRefField);
-					ilgen.Emit(OpCodes.Ldarga_S, (byte)1);
+					ilgen.EmitLdarga(1);
 					ilgen.Emit(OpCodes.Ldfld, ghostRefField);
 					ilgen.Emit(OpCodes.Ceq);
 					ilgen.Emit(OpCodes.Ldc_I4_0);
@@ -1132,7 +1136,7 @@ namespace IKVM.Internal
 					rank++;
 					tw = tw.ElementTypeWrapper;
 				}
-				ilgen.Emit(OpCodes.Ldc_I4, rank);
+				ilgen.EmitLdc_I4(rank);
 				ilgen.Emit(OpCodes.Call, ghostCastArrayMethod);
 				ilgen.Emit(OpCodes.Castclass, ArrayTypeWrapper.MakeArrayType(Types.Object, rank));
 			}
@@ -1171,6 +1175,19 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal void SetEnumType(Type enumType)
+		{
+			this.enumType = enumType;
+		}
+
+		internal override Type EnumType
+		{
+			get
+			{
+				return enumType;
+			}
+		}
+
 		private sealed class ReplacedMethodWrapper : MethodWrapper
 		{
 			private IKVM.Internal.MapXml.InstructionList code;
