@@ -34,7 +34,7 @@ using IKVM.Internal;
 
 namespace IKVM.Internal.MapXml
 {
-	class CodeGenContext
+	sealed class CodeGenContext
 	{
 		private ClassLoaderWrapper classLoader;
 		private readonly Dictionary<string, object> h = new Dictionary<string, object>();
@@ -377,7 +377,7 @@ namespace IKVM.Internal.MapXml
 					// typeWrapper should never be an UnloadableTypeWrapper
 					typeWrapper.EmitInstanceOf(null, ilgen);
 					CodeEmitterLabel endLabel = ilgen.DefineLabel();
-					ilgen.Emit(OpCodes.Brtrue_S, endLabel);
+					ilgen.EmitBrtrue(endLabel);
 					ilgen.Emit(OpCodes.Pop);
 					ilgen.Emit(OpCodes.Ldnull);
 					ilgen.MarkLabel(endLabel);
@@ -479,13 +479,6 @@ namespace IKVM.Internal.MapXml
 
 	public abstract class Branch : Instruction
 	{
-		private OpCode opcode;
-
-		public Branch(OpCode opcode)
-		{
-			this.opcode = opcode;
-		}
-
 		internal sealed override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
 			CodeEmitterLabel l;
@@ -498,9 +491,11 @@ namespace IKVM.Internal.MapXml
 			{
 				l = (CodeEmitterLabel)context[Name];
 			}
-			ilgen.Emit(opcode, l);
+			Emit(ilgen, l);
 		}
 
+		internal abstract void Emit(CodeEmitter ilgen, CodeEmitterLabel label);
+
 		[XmlAttribute("name")]
 		public string Name;
 	}
@@ -508,74 +503,81 @@ namespace IKVM.Internal.MapXml
 	[XmlType("brfalse")]
 	public sealed class BrFalse : Branch
 	{
-		public BrFalse() : base(OpCodes.Brfalse)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBrfalse(label);
 		}
 	}
 
 	[XmlType("brtrue")]
 	public sealed class BrTrue : Branch
 	{
-		public BrTrue() : base(OpCodes.Brtrue)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBrtrue(label);
 		}
 	}
 
 	[XmlType("br")]
 	public sealed class Br : Branch
 	{
-		public Br() : base(OpCodes.Br)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBr(label);
 		}
 	}
 
 	[XmlType("beq")]
 	public sealed class Beq : Branch
 	{
-		public Beq()
-			: base(OpCodes.Beq)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBeq(label);
 		}
 	}
 
 	[XmlType("bne_un")]
 	public sealed class Bne_Un : Branch
 	{
-		public Bne_Un()
-			: base(OpCodes.Bne_Un)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBne_Un(label);
 		}
 	}
 
 	[XmlType("bge_un")]
 	public sealed class Bge_Un : Branch
 	{
-		public Bge_Un() : base(OpCodes.Bge_Un)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBge_Un(label);
 		}
 	}
 
 	[XmlType("ble_un")]
 	public sealed class Ble_Un : Branch
 	{
-		public Ble_Un() : base(OpCodes.Ble_Un)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBle_Un(label);
 		}
 	}
 
 	[XmlType("blt")]
 	public sealed class Blt : Branch
 	{
-		public Blt() : base(OpCodes.Blt)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBlt(label);
 		}
 	}
 
 	[XmlType("blt_un")]
 	public sealed class Blt_Un : Branch
 	{
-		public Blt_Un() : base(OpCodes.Blt_Un)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitBlt_Un(label);
 		}
 	}
 
@@ -658,7 +660,7 @@ namespace IKVM.Internal.MapXml
 
 		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Ldarga, (short)ArgNum);
+			ilgen.EmitLdarga(ArgNum);
 		}
 	}
 
@@ -670,7 +672,7 @@ namespace IKVM.Internal.MapXml
 
 		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Ldarg_S, ArgNum);
+			ilgen.EmitLdarg(ArgNum);
 		}
 	}
 
@@ -930,7 +932,7 @@ namespace IKVM.Internal.MapXml
 
 		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Ldc_I4, val);
+			ilgen.EmitLdc_I4(val);
 		}
 	}
 
@@ -1107,7 +1109,7 @@ namespace IKVM.Internal.MapXml
 
 		internal override void Generate(CodeGenContext context, CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Unaligned, (byte)Alignment);
+			ilgen.EmitUnaligned((byte)Alignment);
 		}
 	}
 
@@ -1130,8 +1132,9 @@ namespace IKVM.Internal.MapXml
 	[XmlType("leave")]
 	public sealed class Leave : Branch
 	{
-		public Leave() : base(OpCodes.Leave)
+		internal override void Emit(CodeEmitter ilgen, CodeEmitterLabel label)
 		{
+			ilgen.EmitLeave(label);
 		}
 	}
 
@@ -1177,7 +1180,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class CatchBlock : InstructionList
+	public sealed class CatchBlock : InstructionList
 	{
 		[XmlAttribute("type")]
 		public string type;
@@ -1186,7 +1189,7 @@ namespace IKVM.Internal.MapXml
 	}
 
 	[XmlType("conditional")]
-	public class ConditionalInstruction : Instruction
+	public sealed class ConditionalInstruction : Instruction
 	{
 		[XmlAttribute("framework")]
 		public string framework;
@@ -1499,13 +1502,13 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class Throws
+	public sealed class Throws
 	{
 		[XmlAttribute("class")]
 		public string Class;
 	}
 
-	public class Redirect
+	public sealed class Redirect
 	{
 		private int linenum = Root.LineNumber;
 
@@ -1535,7 +1538,7 @@ namespace IKVM.Internal.MapXml
 			Type[] redirParamTypes = loader.ArgTypeListFromSig(Sig);
 			for(int i = 0; i < redirParamTypes.Length; i++)
 			{
-				ilgen.Emit(OpCodes.Ldarg, (short)i);
+				ilgen.EmitLdarg(i);
 			}
 			// HACK if the class name contains a comma, we assume it is a .NET type
 			if(Class.IndexOf(',') >= 0)
@@ -1564,7 +1567,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class Override
+	public sealed class Override
 	{
 		[XmlAttribute("class")]
 		public string Class;
@@ -1572,7 +1575,7 @@ namespace IKVM.Internal.MapXml
 		public string Name;
 	}
 
-	public class ReplaceMethodCall
+	public sealed class ReplaceMethodCall
 	{
 		[XmlAttribute("class")]
 		public string Class;
@@ -1627,7 +1630,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class Method : MethodConstructorBase
+	public sealed class Method : MethodConstructorBase
 	{
 		[XmlAttribute("name")]
 		public string Name;
@@ -1642,7 +1645,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class Constructor : MethodConstructorBase
+	public sealed class Constructor : MethodConstructorBase
 	{
 		internal override MethodKey ToMethodKey(string className)
 		{
@@ -1650,7 +1653,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class ClassInitializer : MethodBase
+	public sealed class ClassInitializer : MethodBase
 	{
 		internal override MethodKey ToMethodKey(string className)
 		{
@@ -1658,7 +1661,7 @@ namespace IKVM.Internal.MapXml
 		}
 	}
 
-	public class Field
+	public sealed class Field
 	{
 		[XmlAttribute("name")]
 		public string Name;
@@ -1668,12 +1671,11 @@ namespace IKVM.Internal.MapXml
 		public MapModifiers Modifiers;
 		[XmlAttribute("constant")]
 		public string Constant;
-		public Redirect redirect;
 		[XmlElement("attribute")]
 		public Attribute[] Attributes;
 	}
 
-	public class Property
+	public sealed class Property
 	{
 		[XmlAttribute("name")]
 		public string Name;
@@ -1685,7 +1687,7 @@ namespace IKVM.Internal.MapXml
 		public Attribute[] Attributes;
 	}
 
-	public class Interface
+	public sealed class Interface
 	{
 		[XmlAttribute("class")]
 		public string Name;
@@ -1709,7 +1711,11 @@ namespace IKVM.Internal.MapXml
 		[XmlEnum("static")]
 		Static = Modifiers.Static,
 		[XmlEnum("abstract")]
-		Abstract = Modifiers.Abstract
+		Abstract = Modifiers.Abstract,
+		[XmlEnum("ACC_BRIDGE")]
+		Bridge = Modifiers.Bridge,
+		[XmlEnum("ACC_SYNTHETIC")]
+		Synthetic = Modifiers.Synthetic,
 	}
 
 	public enum Scope
@@ -1720,13 +1726,13 @@ namespace IKVM.Internal.MapXml
 		Private = 1
 	}
 
-	public class Element
+	public sealed class Element
 	{
 		[XmlText]
 		public string Value;
 	}
 
-	public class Param
+	public sealed class Param
 	{
 		[XmlText]
 		public string Value;
@@ -1740,7 +1746,7 @@ namespace IKVM.Internal.MapXml
 		public Attribute[] Attributes;
 	}
 
-	public class Attribute
+	public sealed class Attribute
 	{
 		[XmlAttribute("type")]
 		public string Type;
@@ -1757,7 +1763,7 @@ namespace IKVM.Internal.MapXml
 	}
 
 	[XmlType("class")]
-	public class Class
+	public sealed class Class
 	{
 		[XmlAttribute("name")]
 		public string Name;
@@ -1783,7 +1789,7 @@ namespace IKVM.Internal.MapXml
 		public Attribute[] Attributes;
 	}
 
-	public class Assembly
+	public sealed class Assembly
 	{
 		[XmlElement("class")]
 		public Class[] Classes;
@@ -1792,7 +1798,7 @@ namespace IKVM.Internal.MapXml
 	}
 
 	[XmlType("exception")]
-	public class ExceptionMapping
+	public sealed class ExceptionMapping
 	{
 		[XmlAttribute]
 		public string src;
@@ -1802,10 +1808,9 @@ namespace IKVM.Internal.MapXml
 	}
 
 	[XmlRoot("root")]
-	public class Root
+	public sealed class Root
 	{
 		internal static System.Xml.XmlTextReader xmlReader;
-		internal static string filename;
 
 		internal static int LineNumber
 		{
