@@ -25,6 +25,7 @@
 
 package java.awt.image;
 
+import cli.System.Drawing.Color;
 import cli.System.Drawing.Imaging.*;
 
 import java.awt.Transparency;
@@ -686,12 +687,14 @@ public class BufferedImage extends java.awt.Image
         {
             throw new IllegalArgumentException();
         }
-        bitmap = createBitmap(width, height);
-        cli.System.Drawing.Rectangle rect = new cli.System.Drawing.Rectangle(0, 0, width, height);
-        cli.System.Drawing.Imaging.BitmapData data = bitmap.LockBits(rect, ImageLockMode.wrap(ImageLockMode.WriteOnly), PixelFormat.wrap(PixelFormat.Format32bppArgb));
-        cli.System.IntPtr pixelPtr = data.get_Scan0();
-        cli.System.Runtime.InteropServices.Marshal.Copy(pixelData, 0, pixelPtr, (int)size);
-        bitmap.UnlockBits(data);
+        synchronized( bitmap ) {            
+            bitmap = createBitmap(width, height);
+            cli.System.Drawing.Rectangle rect = new cli.System.Drawing.Rectangle(0, 0, width, height);
+            cli.System.Drawing.Imaging.BitmapData data = bitmap.LockBits(rect, ImageLockMode.wrap(ImageLockMode.WriteOnly), PixelFormat.wrap(PixelFormat.Format32bppArgb));
+            cli.System.IntPtr pixelPtr = data.get_Scan0();
+            cli.System.Runtime.InteropServices.Marshal.Copy(pixelData, 0, pixelPtr, (int)size);
+            bitmap.UnlockBits(data);
+        }
     }
 
     /**
@@ -703,28 +706,30 @@ public class BufferedImage extends java.awt.Image
         if(currentBuffer != BUFFER_BITMAP){
             return; // BUFFER_BOTH and BUFFER_RASTER
         }
-        int width = bitmap.get_Width();
-        int height = bitmap.get_Height();
-        if(colorModel == null){
-            colorModel = createColorModel();
-        }
-        if(raster == null){
-            raster = createRaster(width, height);
-        }
-        
-        switch (getType()){
-            case TYPE_INT_ARGB:
-                copyFromBitmap(bitmap, ((DataBufferInt)raster.getDataBuffer()).getData());
-                break;
-            default:
-                for( int y = 0; y<height; y++){
-                    for(int x = 0; x<width; x++){
-                    	int rgb = bitmap.GetPixel(x, y).ToArgb();
-                    	raster.setDataElements(x, y, colorModel.getDataElements(rgb, null));
+        synchronized( bitmap ) {
+            int width = bitmap.get_Width();
+            int height = bitmap.get_Height();
+            if(colorModel == null){
+                colorModel = createColorModel();
+            }
+            if(raster == null){
+                raster = createRaster(width, height);
+            }
+            
+            switch (getType()){
+                case TYPE_INT_ARGB:
+                    copyFromBitmap(bitmap, ((DataBufferInt)raster.getDataBuffer()).getData());
+                    break;
+                default:
+                    for( int y = 0; y<height; y++){
+                        for(int x = 0; x<width; x++){
+                            int rgb = bitmap.GetPixel(x, y).ToArgb();
+                            raster.setDataElements(x, y, colorModel.getDataElements(rgb, null));
+                        }
                     }
-                }
+            }
+            this.currentBuffer = BUFFER_BOTH;
         }
-        this.currentBuffer = BUFFER_BOTH;
     }
 
     @cli.System.Security.SecuritySafeCriticalAttribute.Annotation
@@ -1073,9 +1078,22 @@ public class BufferedImage extends java.awt.Image
      * @see #getRGB(int, int, int, int, int[], int, int)
      */
     public synchronized void setRGB(int x, int y, int rgb) {
-        bitmap2Raster();
-        this.currentBuffer = BUFFER_RASTER;
-        raster.setDataElements(x, y, colorModel.getDataElements(rgb, null));
+    	switch( currentBuffer ) {
+    	case BUFFER_BITMAP:
+    		synchronized (bitmap) {
+    			bitmap.SetPixel(x, y, Color.FromArgb(rgb));
+			}
+    		break;
+    	case BUFFER_RASTER:
+    		raster.setDataElements(x, y, colorModel.getDataElements(rgb, null));
+    		break;
+    	case BUFFER_BOTH:
+    		raster.setDataElements(x, y, colorModel.getDataElements(rgb, null));
+    		synchronized (bitmap) {
+    			bitmap.SetPixel(x, y, Color.FromArgb(rgb));
+			}
+    		break;
+    	}
     }
 
     /**
@@ -1110,17 +1128,13 @@ public class BufferedImage extends java.awt.Image
      */
     public void setRGB(int startX, int startY, int w, int h,
                         int[] rgbArray, int offset, int scansize) {
-        bitmap2Raster();
-        this.currentBuffer = BUFFER_RASTER;
         int yoff  = offset;
         int off;
-        Object pixel = null;
 
         for (int y = startY; y < startY+h; y++, yoff+=scansize) {
             off = yoff;
             for (int x = startX; x < startX+w; x++) {
-                pixel = colorModel.getDataElements(rgbArray[off++], pixel);
-                raster.setDataElements(x, y, pixel);
+            	setRGB(x, y, rgbArray[off++]);
             }
         }
     }
@@ -1131,10 +1145,12 @@ public class BufferedImage extends java.awt.Image
      * @return the width of this <code>BufferedImage</code>
      */
     public int getWidth() {
-        if(currentBuffer == BUFFER_RASTER){
-            return raster.getWidth();
+        if(currentBuffer == BUFFER_BITMAP){
+            synchronized( bitmap ) {
+                return bitmap.get_Width();
+            }
         }else{
-            return bitmap.get_Width();
+            return raster.getWidth();
         }
     }
 
@@ -1143,10 +1159,12 @@ public class BufferedImage extends java.awt.Image
      * @return the height of this <code>BufferedImage</code>
      */
     public int getHeight() {
-        if(currentBuffer == BUFFER_RASTER){
-            return raster.getHeight();
+        if(currentBuffer == BUFFER_BITMAP){
+            synchronized( bitmap ) {
+                return bitmap.get_Height();
+            }
         }else{
-            return bitmap.get_Height();
+            return raster.getHeight();
         }
     }
 
@@ -1176,15 +1194,17 @@ public class BufferedImage extends java.awt.Image
      */
     public ImageProducer getSource(){
         if(currentBuffer != BUFFER_RASTER){
-            int width = bitmap.get_Width();
-            int height = bitmap.get_Height();
-            int[] pix = new int[width * height];
-            for(int y = 0; y < height; y++){
-                for(int x = 0; x < width; x++){
-                    pix[x + y * width] = bitmap.GetPixel(x, y).ToArgb();
+            synchronized( bitmap ) {
+                int width = bitmap.get_Width();
+                int height = bitmap.get_Height();
+                int[] pix = new int[width * height];
+                for(int y = 0; y < height; y++){
+                    for(int x = 0; x < width; x++){
+                        pix[x + y * width] = bitmap.GetPixel(x, y).ToArgb();
+                    }
                 }
+                return new java.awt.image.MemoryImageSource(width, height, pix, 0, width);
             }
-            return new java.awt.image.MemoryImageSource(width, height, pix, 0, width);
         }else{
             if(osis == null){
                 if(properties == null){
@@ -1261,7 +1281,7 @@ public class BufferedImage extends java.awt.Image
      *          image.
      */
     public Graphics2D createGraphics() {
-        ikvm.awt.IkvmToolkit toolkit = (ikvm.awt.IkvmToolkit)java.awt.Toolkit.getDefaultToolkit();
+        ikvm.awt.IkvmToolkit toolkit = ikvm.awt.IkvmToolkit.DefaultToolkit.get();
         raster2Bitmap();
         this.currentBuffer = BUFFER_BITMAP;
         return toolkit.createGraphics( bitmap );
