@@ -1903,6 +1903,17 @@ public abstract class ClassLoader {
     }
 
     private static boolean loadLibrary0(Class fromClass, final File file) {
+        if (loadLibrary1(fromClass, file)) {
+            return true;
+        }
+        final File libfile = ClassLoaderHelper.mapAlternativeName(file);
+        if (libfile != null && loadLibrary1(fromClass, libfile)) {
+            return true;
+        }
+        return false;
+    }
+
+    private static boolean loadLibrary1(Class fromClass, final File file) {
         boolean exists = AccessController.doPrivileged(
             new PrivilegedAction<Object>() {
                 public Object run() {
@@ -2219,6 +2230,16 @@ public abstract class ClassLoader {
 
     // Retrieves the assertion directives from the VM.
     private static native AssertionStatusDirectives retrieveDirectives();
+
+    // [IKVM] equivalent of HotSpot's java_lang_ClassLoader::is_trusted_loader()
+    static boolean isTrustedLoader(ClassLoader loader) {
+        ClassLoader cl = scl;
+        while (cl != null) {
+            if (cl == loader) return true;
+            cl = cl.parent;
+        }
+        return false;
+    }
 }
 
 
