@@ -129,6 +129,17 @@ public class ZipFile implements ZipConstants, Closeable
     this(file, OPEN_READ, charset);
   }
 
+  /**
+   * Opens a Zip file reading the given File.
+   * @exception IOException if a i/o error occured.
+   * @exception ZipException if the file doesn't contain a valid zip
+   * archive.  
+   */
+  public ZipFile(String name, Charset charset) throws IOException
+  {
+    this(new File(name), OPEN_READ, charset);
+  }
+
   /**
    * Opens a Zip file reading the given File in the given mode.
    *
