@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2006-2011 Jeroen Frijters
+  Copyright (C) 2006-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -56,6 +56,12 @@ public final class Unsafe
         return theUnsafe;
     }
 
+    // this is the intrinsified version of objectFieldOffset(XXX.class.getDeclaredField("xxx"))
+    public long objectFieldOffset(Class c, String field)
+    {
+        return fieldOffset(ReflectHelper.createFieldAndMakeAccessible(c, field));
+    }
+
     // NOTE we have a really lame (and slow) implementation!
     public long objectFieldOffset(Field field)
     {
@@ -80,11 +86,13 @@ public final class Unsafe
 
     public int arrayBaseOffset(Class c)
     {
+        // don't change this, the Unsafe intrinsics depend on this value
         return 0;
     }
 
     public int arrayIndexScale(Class c)
     {
+        // don't change this, the Unsafe intrinsics depend on this value
         return 1;
     }
 
@@ -1106,6 +1114,18 @@ public final class Unsafe
         return -1;
     }
 
+    public void park(boolean isAbsolute, long time)
+    {
+        if (isAbsolute)
+        {
+            java.util.concurrent.locks.LockSupport.parkUntil(time);
+        }
+        else
+        {
+            java.util.concurrent.locks.LockSupport.parkNanos(time);
+        }
+    }
+
     public void unpark(Object thread)
     {
         java.util.concurrent.locks.LockSupport.unpark((Thread)thread);
@@ -1117,4 +1137,19 @@ public final class Unsafe
     }
 
     public native Class defineClass(String name, byte[] buf, int offset, int length, ClassLoader cl, ProtectionDomain pd);
+
+    public void monitorEnter(Object o)
+    {
+        cli.System.Threading.Monitor.Enter(o);
+    }
+
+    public void monitorExit(Object o)
+    {
+        cli.System.Threading.Monitor.Exit(o);
+    }
+
+    public boolean tryMonitorEnter(Object o)
+    {
+        return cli.System.Threading.Monitor.TryEnter(o);
+    }
 }
