@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2008-2010 Jeroen Frijters
+  Copyright (C) 2008-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -26,7 +26,6 @@ using System.Runtime.InteropServices;
 using System.Collections.Generic;
 using System.Diagnostics.SymbolStore;
 using System.Diagnostics;
-using IKVM.Reflection.Metadata;
 using IKVM.Reflection.Writer;
 
 namespace IKVM.Reflection.Emit
@@ -72,20 +71,20 @@ namespace IKVM.Reflection.Emit
 		}
 	}
 
-	public sealed class LocalBuilder
+	public sealed class LocalBuilder : LocalVariableInfo
 	{
-		private readonly Type localType;
-		private readonly int index;
-		private readonly bool pinned;
 		internal string name;
 		internal int startOffset;
 		internal int endOffset;
 
 		internal LocalBuilder(Type localType, int index, bool pinned)
+			: base(index, localType, pinned)
+		{
+		}
+
+		internal LocalBuilder(Type localType, int index, bool pinned, CustomModifiers customModifiers)
+			: base(index, localType, pinned, customModifiers)
 		{
-			this.localType = localType;
-			this.index = index;
-			this.pinned = pinned;
 		}
 
 		public void SetLocalSymInfo(string name)
@@ -99,59 +98,14 @@ namespace IKVM.Reflection.Emit
 			this.startOffset = startOffset;
 			this.endOffset = endOffset;
 		}
-
-		public Type LocalType
-		{
-			get { return localType; }
-		}
-
-		public int LocalIndex
-		{
-			get { return index; }
-		}
-
-		public bool IsPinned
-		{
-			get { return pinned; }
-		}
-	}
-
-	sealed class MarkerType : Type
-	{
-		public override Type BaseType
-		{
-			get { throw new InvalidOperationException(); }
-		}
-
-		public override TypeAttributes Attributes
-		{
-			get { throw new InvalidOperationException(); }
-		}
-
-		public override string Name
-		{
-			get { throw new InvalidOperationException(); }
-		}
-
-		public override string FullName
-		{
-			get { throw new InvalidOperationException(); }
-		}
-
-		public override Module Module
-		{
-			get { throw new InvalidOperationException(); }
-		}
 	}
 
 	public sealed class ILGenerator
 	{
-		private static readonly Type FAULT = new MarkerType();
-		private static readonly Type FINALLY = new MarkerType();
-		private static readonly Type FILTER = new MarkerType();
 		private readonly ModuleBuilder moduleBuilder;
 		private readonly ByteBuffer code;
-		private readonly List<LocalBuilder> locals = new List<LocalBuilder>();
+		private readonly SignatureHelper locals;
+		private int localsCount;
 		private readonly List<int> tokenFixups = new List<int>();
 		private readonly List<int> labels = new List<int>();
 		private readonly List<int> labelStackHeight = new List<int>();
@@ -182,7 +136,7 @@ namespace IKVM.Reflection.Emit
 			internal int tryLength;
 			internal int handlerOffset;
 			internal int handlerLength;
-			internal Type exceptionType;	// FINALLY = finally block, FILTER = handler with filter, FAULT = fault block
+			internal Type exceptionType;	// MarkerType.Finally = finally block, MarkerType.Filter = handler with filter, MarkerType.Fault = fault block
 			internal int filterOffset;
 
 			internal ExceptionBlock(int ordinal)
@@ -244,17 +198,13 @@ namespace IKVM.Reflection.Emit
 		{
 			this.code = new ByteBuffer(initialCapacity);
 			this.moduleBuilder = moduleBuilder;
+			this.locals = SignatureHelper.GetLocalVarSigHelper(moduleBuilder);
 			if (moduleBuilder.symbolWriter != null)
 			{
 				scope = new Scope(null);
 			}
 		}
 
-		private bool IsLabelReachable(Label label)
-		{
-			return labelStackHeight[label.Index] != -1;
-		}
-
 		// non-standard API
 		public void __DisableExceptionBlockAssistance()
 		{
@@ -309,7 +259,7 @@ namespace IKVM.Reflection.Emit
 			UpdateStack(1);
 			if (exceptionType == null)
 			{
-				if (block.exceptionType != FILTER || block.handlerOffset != 0)
+				if (block.exceptionType != MarkerType.Filter || block.handlerOffset != 0)
 				{
 					throw new ArgumentNullException("exceptionType");
 				}
@@ -334,7 +284,7 @@ namespace IKVM.Reflection.Emit
 					exceptionStack.Push(block);
 				}
 				block.exceptionType = exceptionType;
-				if (exceptionType == FILTER)
+				if (exceptionType == MarkerType.Filter)
 				{
 					block.filterOffset = code.Position;
 				}
@@ -358,17 +308,17 @@ namespace IKVM.Reflection.Emit
 
 		public void BeginExceptFilterBlock()
 		{
-			BeginCatchBlock(FILTER);
+			BeginCatchBlock(MarkerType.Filter);
 		}
 
 		public void BeginFaultBlock()
 		{
-			BeginFinallyFaultBlock(FAULT);
+			BeginFinallyFaultBlock(MarkerType.Fault);
 		}
 
 		public void BeginFinallyBlock()
 		{
-			BeginFinallyFaultBlock(FINALLY);
+			BeginFinallyFaultBlock(MarkerType.Finally);
 		}
 
 		private void BeginFinallyFaultBlock(Type type)
@@ -415,7 +365,7 @@ namespace IKVM.Reflection.Emit
 			ExceptionBlock block = exceptionStack.Pop();
 			if (exceptionBlockAssistanceMode == EBAM_COMPAT || (exceptionBlockAssistanceMode == EBAM_CLEVER && stackHeight != -1))
 			{
-				if (block.filterOffset != 0 || (block.exceptionType != FINALLY && block.exceptionType != FAULT))
+				if (block.filterOffset != 0 || (block.exceptionType != MarkerType.Finally && block.exceptionType != MarkerType.Fault))
 				{
 					Emit(OpCodes.Leave, block.labelEnd);
 				}
@@ -451,8 +401,19 @@ namespace IKVM.Reflection.Emit
 
 		public LocalBuilder DeclareLocal(Type localType, bool pinned)
 		{
-			LocalBuilder local = new LocalBuilder(localType, locals.Count, pinned);
-			locals.Add(local);
+			LocalBuilder local = new LocalBuilder(localType, localsCount++, pinned);
+			locals.AddArgument(localType, pinned);
+			if (scope != null)
+			{
+				scope.locals.Add(local);
+			}
+			return local;
+		}
+
+		public LocalBuilder __DeclareLocal(Type localType, bool pinned, CustomModifiers customModifiers)
+		{
+			LocalBuilder local = new LocalBuilder(localType, localsCount++, pinned);
+			locals.__AddArgument(localType, pinned, customModifiers);
 			if (scope != null)
 			{
 				scope.locals.Add(local);
@@ -765,7 +726,7 @@ namespace IKVM.Reflection.Emit
 		public void Emit(OpCode opc, string str)
 		{
 			Emit(opc);
-			code.Write(0x70000000 | moduleBuilder.UserStrings.Add(str));
+			code.Write(moduleBuilder.GetStringConstant(str).Token);
 		}
 
 		public void Emit(OpCode opc, Type type)
@@ -785,10 +746,15 @@ namespace IKVM.Reflection.Emit
 		{
 			Emit(opcode);
 			UpdateStack(opcode, signature.HasThis, signature.ReturnType, signature.ParameterCount);
-			code.Write(0x11000000 | moduleBuilder.StandAloneSig.FindOrAddRecord(moduleBuilder.Blobs.Add(signature.GetSignature(moduleBuilder))));
+			code.Write(moduleBuilder.GetSignatureToken(signature).Token);
 		}
 
 		public void EmitCall(OpCode opc, MethodInfo method, Type[] optionalParameterTypes)
+		{
+			__EmitCall(opc, method, optionalParameterTypes, null);
+		}
+
+		public void __EmitCall(OpCode opc, MethodInfo method, Type[] optionalParameterTypes, CustomModifiers[] customModifiers)
 		{
 			if (optionalParameterTypes == null || optionalParameterTypes.Length == 0)
 			{
@@ -798,20 +764,7 @@ namespace IKVM.Reflection.Emit
 			{
 				Emit(opc);
 				UpdateStack(opc, method.HasThis, method.ReturnType, method.ParameterCount + optionalParameterTypes.Length);
-				ByteBuffer sig = new ByteBuffer(16);
-				method.MethodSignature.WriteMethodRefSig(moduleBuilder, sig, optionalParameterTypes);
-				MemberRefTable.Record record = new MemberRefTable.Record();
-				if (method.Module == moduleBuilder)
-				{
-					record.Class = method.MetadataToken;
-				}
-				else
-				{
-					record.Class = moduleBuilder.GetTypeTokenForMemberRef(method.DeclaringType ?? method.Module.GetModuleType());
-				}
-				record.Name = moduleBuilder.Strings.Add(method.Name);
-				record.Signature = moduleBuilder.Blobs.Add(sig);
-				code.Write(0x0A000000 | moduleBuilder.MemberRef.FindOrAddRecord(record));
+				code.Write(moduleBuilder.__GetMethodToken(method, optionalParameterTypes, customModifiers).Token);
 			}
 		}
 
@@ -820,25 +773,42 @@ namespace IKVM.Reflection.Emit
 			EmitCall(opc, constructor.GetMethodInfo(), optionalParameterTypes);
 		}
 
+		public void __EmitCall(OpCode opc, ConstructorInfo constructor, Type[] optionalParameterTypes, CustomModifiers[] customModifiers)
+		{
+			__EmitCall(opc, constructor.GetMethodInfo(), optionalParameterTypes, customModifiers);
+		}
+
 		public void EmitCalli(OpCode opc, CallingConvention callingConvention, Type returnType, Type[] parameterTypes)
 		{
-			returnType = returnType ?? moduleBuilder.universe.System_Void;
-			Emit(opc);
-			UpdateStack(opc, false, returnType, parameterTypes.Length);
-			ByteBuffer sig = new ByteBuffer(16);
-			Signature.WriteStandAloneMethodSig(moduleBuilder, sig, callingConvention, returnType, parameterTypes);
-			code.Write(0x11000000 | moduleBuilder.StandAloneSig.FindOrAddRecord(moduleBuilder.Blobs.Add(sig)));
+			SignatureHelper sig = SignatureHelper.GetMethodSigHelper(moduleBuilder, callingConvention, returnType);
+			sig.AddArguments(parameterTypes, null, null);
+			Emit(opc, sig);
 		}
 
 		public void EmitCalli(OpCode opc, CallingConventions callingConvention, Type returnType, Type[] parameterTypes, Type[] optionalParameterTypes)
 		{
-			returnType = returnType ?? moduleBuilder.universe.System_Void;
-			optionalParameterTypes = optionalParameterTypes ?? Type.EmptyTypes;
+			SignatureHelper sig = SignatureHelper.GetMethodSigHelper(moduleBuilder, callingConvention, returnType);
+			sig.AddArguments(parameterTypes, null, null);
+			sig.AddSentinel();
+			sig.AddArguments(optionalParameterTypes, null, null);
+			Emit(opc, sig);
+		}
+
+		public void __EmitCalli(OpCode opc, __StandAloneMethodSig sig)
+		{
 			Emit(opc);
-			UpdateStack(opc, (callingConvention & CallingConventions.HasThis | CallingConventions.ExplicitThis) == CallingConventions.HasThis, returnType, parameterTypes.Length + optionalParameterTypes.Length);
-			ByteBuffer sig = new ByteBuffer(16);
-			Signature.WriteStandAloneMethodSig(moduleBuilder, sig, callingConvention, returnType, parameterTypes, optionalParameterTypes);
-			code.Write(0x11000000 | moduleBuilder.StandAloneSig.FindOrAddRecord(moduleBuilder.Blobs.Add(sig)));
+			if (sig.IsUnmanaged)
+			{
+				UpdateStack(opc, false, sig.ReturnType, sig.ParameterCount);
+			}
+			else
+			{
+				CallingConventions callingConvention = sig.CallingConvention;
+				UpdateStack(opc, (callingConvention & CallingConventions.HasThis | CallingConventions.ExplicitThis) == CallingConventions.HasThis, sig.ReturnType, sig.ParameterCount);
+			}
+			ByteBuffer bb = new ByteBuffer(16);
+			Signature.WriteStandAloneMethodSig(moduleBuilder, bb, sig);
+			code.Write(0x11000000 | moduleBuilder.StandAloneSig.FindOrAddRecord(moduleBuilder.Blobs.Add(bb)));
 		}
 
 		public void EmitWriteLine(string text)
@@ -937,7 +907,7 @@ namespace IKVM.Reflection.Emit
 			int localVarSigTok = 0;
 
 			int rva;
-			if (locals.Count == 0 && exceptions.Count == 0 && maxStack <= 8 && code.Length < 64 && !fatHeader)
+			if (localsCount == 0 && exceptions.Count == 0 && maxStack <= 8 && code.Length < 64 && !fatHeader)
 			{
 				rva = WriteTinyHeaderAndCode(bb);
 			}
@@ -1025,11 +995,9 @@ namespace IKVM.Reflection.Emit
 			bb.Align(4);
 			int rva = bb.Position;
 
-			if (locals.Count != 0)
+			if (localsCount != 0)
 			{
-				ByteBuffer localVarSig = new ByteBuffer(locals.Count + 2);
-				Signature.WriteLocalVarSig(moduleBuilder, localVarSig, locals);
-				localVarSigTok = 0x11000000 | moduleBuilder.StandAloneSig.FindOrAddRecord(moduleBuilder.Blobs.Add(localVarSig));
+				localVarSigTok = moduleBuilder.GetSignatureToken(locals).Token;
 			}
 
 			const byte CorILMethod_FatFormat = 0x03;
@@ -1087,15 +1055,15 @@ namespace IKVM.Reflection.Emit
 					bb.Write((short)(dataSize >> 8));
 					foreach (ExceptionBlock block in exceptions)
 					{
-						if (block.exceptionType == FAULT)
+						if (block.exceptionType == MarkerType.Fault)
 						{
 							bb.Write((int)COR_ILEXCEPTION_CLAUSE_FAULT);
 						}
-						else if (block.exceptionType == FILTER)
+						else if (block.exceptionType == MarkerType.Filter)
 						{
 							bb.Write((int)COR_ILEXCEPTION_CLAUSE_FILTER);
 						}
-						else if (block.exceptionType == FINALLY)
+						else if (block.exceptionType == MarkerType.Finally)
 						{
 							bb.Write((int)COR_ILEXCEPTION_CLAUSE_FINALLY);
 						}
@@ -1107,7 +1075,7 @@ namespace IKVM.Reflection.Emit
 						bb.Write(block.tryLength);
 						bb.Write(block.handlerOffset);
 						bb.Write(block.handlerLength);
-						if (block.exceptionType != FAULT && block.exceptionType != FILTER && block.exceptionType != FINALLY)
+						if (block.exceptionType != MarkerType.Fault && block.exceptionType != MarkerType.Filter && block.exceptionType != MarkerType.Finally)
 						{
 							bb.Write(moduleBuilder.GetTypeTokenForMemberRef(block.exceptionType));
 						}
@@ -1124,15 +1092,15 @@ namespace IKVM.Reflection.Emit
 					bb.Write((short)0);
 					foreach (ExceptionBlock block in exceptions)
 					{
-						if (block.exceptionType == FAULT)
+						if (block.exceptionType == MarkerType.Fault)
 						{
 							bb.Write(COR_ILEXCEPTION_CLAUSE_FAULT);
 						}
-						else if (block.exceptionType == FILTER)
+						else if (block.exceptionType == MarkerType.Filter)
 						{
 							bb.Write(COR_ILEXCEPTION_CLAUSE_FILTER);
 						}
-						else if (block.exceptionType == FINALLY)
+						else if (block.exceptionType == MarkerType.Finally)
 						{
 							bb.Write(COR_ILEXCEPTION_CLAUSE_FINALLY);
 						}
@@ -1144,7 +1112,7 @@ namespace IKVM.Reflection.Emit
 						bb.Write((byte)block.tryLength);
 						bb.Write((short)block.handlerOffset);
 						bb.Write((byte)block.handlerLength);
-						if (block.exceptionType != FAULT && block.exceptionType != FILTER && block.exceptionType != FINALLY)
+						if (block.exceptionType != MarkerType.Fault && block.exceptionType != MarkerType.Filter && block.exceptionType != MarkerType.Finally)
 						{
 							bb.Write(moduleBuilder.GetTypeTokenForMemberRef(block.exceptionType));
 						}
