@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2009-2011 Jeroen Frijters
+  Copyright (C) 2009-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -98,33 +98,30 @@ namespace IKVM.Reflection
 			{
 				throw new BadImageFormatException();
 			}
-			int genArgCount = br.ReadCompressedInt();
+			int genArgCount = br.ReadCompressedUInt();
 			Type[] args = new Type[genArgCount];
-			Type[][] reqmod = null;
-			Type[][] optmod = null;
+			CustomModifiers[] mods = null;
 			for (int i = 0; i < genArgCount; i++)
 			{
 				// LAMESPEC the Type production (23.2.12) doesn't include CustomMod* for genericinst, but C++ uses it, the verifier allows it and ildasm also supports it
-				CustomModifiers mods = ReadCustomModifiers(module, br, context);
-				if (mods.required != null || mods.optional != null)
+				CustomModifiers cm = CustomModifiers.Read(module, br, context);
+				if (!cm.IsEmpty)
 				{
-					if (reqmod == null)
+					if (mods == null)
 					{
-						reqmod = new Type[genArgCount][];
-						optmod = new Type[genArgCount][];
+						mods = new CustomModifiers[genArgCount];
 					}
-					reqmod[i] = mods.required;
-					optmod[i] = mods.optional;
+					mods[i] = cm;
 				}
 				args[i] = ReadType(module, br, context);
 			}
-			return GenericTypeInstance.Make(type, args, reqmod, optmod);
+			return GenericTypeInstance.Make(type, args, mods);
 		}
 
 		internal static Type ReadTypeSpec(ModuleReader module, ByteReader br, IGenericContext context)
 		{
 			// LAMESPEC a TypeSpec can contain custom modifiers (C++/CLI generates "newarr (TypeSpec with custom modifiers)")
-			SkipCustomModifiers(br);
+			CustomModifiers.Skip(br);
 			// LAMESPEC anything can be adorned by (useless) custom modifiers
 			// also, VAR and MVAR are also used in TypeSpec (contrary to what the spec says)
 			return ReadType(module, br, context);
@@ -132,10 +129,16 @@ namespace IKVM.Reflection
 
 		private static Type ReadFunctionPointer(ModuleReader module, ByteReader br, IGenericContext context)
 		{
-			// TODO like .NET we return System.IntPtr here, but ideally we should fire an event in Universe that
-			// the user can hook to provide a custom type (or we simply should build in full support for function pointer types)
-			MethodSignature.ReadStandAloneMethodSig(module, br, context);
-			return module.universe.System_IntPtr;
+			__StandAloneMethodSig sig = MethodSignature.ReadStandAloneMethodSig(module, br, context);
+			if (module.universe.EnableFunctionPointers)
+			{
+				return FunctionPointerType.Make(module.universe, sig);
+			}
+			else
+			{
+				// by default, like .NET we return System.IntPtr here
+				return module.universe.System_IntPtr;
+			}
 		}
 
 		internal static Type[] ReadMethodSpec(ModuleReader module, ByteReader br, IGenericContext context)
@@ -144,7 +147,7 @@ namespace IKVM.Reflection
 			{
 				throw new BadImageFormatException();
 			}
-			Type[] args = new Type[br.ReadCompressedInt()];
+			Type[] args = new Type[br.ReadCompressedUInt()];
 			for (int i = 0; i < args.Length; i++)
 			{
 				args[i] = ReadType(module, br, context);
@@ -152,9 +155,24 @@ namespace IKVM.Reflection
 			return args;
 		}
 
+		private static int[] ReadArraySizes(ByteReader br)
+		{
+			int num = br.ReadCompressedUInt();
+			if (num == 0)
+			{
+				return null;
+			}
+			int[] arr = new int[num];
+			for (int i = 0; i < num; i++)
+			{
+				arr[i] = br.ReadCompressedUInt();
+			}
+			return arr;
+		}
+
 		private static int[] ReadArrayBounds(ByteReader br)
 		{
-			int num = br.ReadCompressedInt();
+			int num = br.ReadCompressedUInt();
 			if (num == 0)
 			{
 				return null;
@@ -223,20 +241,20 @@ namespace IKVM.Reflection
 				case ELEMENT_TYPE_OBJECT:
 					return module.universe.System_Object;
 				case ELEMENT_TYPE_VAR:
-					return context.GetGenericTypeArgument(br.ReadCompressedInt());
+					return context.GetGenericTypeArgument(br.ReadCompressedUInt());
 				case ELEMENT_TYPE_MVAR:
-					return context.GetGenericMethodArgument(br.ReadCompressedInt());
+					return context.GetGenericMethodArgument(br.ReadCompressedUInt());
 				case ELEMENT_TYPE_GENERICINST:
 					return ReadGenericInst(module, br, context);
 				case ELEMENT_TYPE_SZARRAY:
-					mods = ReadCustomModifiers(module, br, context);
-					return ReadType(module, br, context).__MakeArrayType(mods.required, mods.optional);
+					mods = CustomModifiers.Read(module, br, context);
+					return ReadType(module, br, context).__MakeArrayType(mods);
 				case ELEMENT_TYPE_ARRAY:
-					mods = ReadCustomModifiers(module, br, context);
-					return ReadType(module, br, context).__MakeArrayType(br.ReadCompressedInt(), ReadArrayBounds(br), ReadArrayBounds(br), mods.required, mods.optional);
+					mods = CustomModifiers.Read(module, br, context);
+					return ReadType(module, br, context).__MakeArrayType(br.ReadCompressedUInt(), ReadArraySizes(br), ReadArrayBounds(br), mods);
 				case ELEMENT_TYPE_PTR:
-					mods = ReadCustomModifiers(module, br, context);
-					return ReadTypeOrVoid(module, br, context).__MakePointerType(mods.required, mods.optional);
+					mods = CustomModifiers.Read(module, br, context);
+					return ReadTypeOrVoid(module, br, context).__MakePointerType(mods);
 				case ELEMENT_TYPE_FNPTR:
 					return ReadFunctionPointer(module, br, context);
 				default:
@@ -250,26 +268,26 @@ namespace IKVM.Reflection
 			{
 				throw new BadImageFormatException("Invalid local variable signature");
 			}
-			int count = br.ReadCompressedInt();
+			int count = br.ReadCompressedUInt();
 			for (int i = 0; i < count; i++)
 			{
 				if (br.PeekByte() == ELEMENT_TYPE_TYPEDBYREF)
 				{
 					br.ReadByte();
-					list.Add(new LocalVariableInfo(i, module.universe.System_TypedReference, false));
+					list.Add(new LocalVariableInfo(i, module.universe.System_TypedReference, false, new CustomModifiers()));
 				}
 				else
 				{
-					SkipCustomModifiers(br);
+					CustomModifiers mods1 = CustomModifiers.Read(module, br, context);
 					bool pinned = false;
 					if (br.PeekByte() == ELEMENT_TYPE_PINNED)
 					{
 						br.ReadByte();
 						pinned = true;
 					}
-					SkipCustomModifiers(br);
+					CustomModifiers mods2 = CustomModifiers.Read(module, br, context);
 					Type type = ReadTypeOrByRef(module, br, context);
-					list.Add(new LocalVariableInfo(i, type, pinned));
+					list.Add(new LocalVariableInfo(i, type, pinned, CustomModifiers.Combine(mods1, mods2)));
 				}
 			}
 		}
@@ -281,9 +299,9 @@ namespace IKVM.Reflection
 				br.ReadByte();
 				// LAMESPEC it is allowed (by C++/CLI, ilasm and peverify) to have custom modifiers after the BYREF
 				// (which makes sense, as it is analogous to pointers)
-				CustomModifiers mods = ReadCustomModifiers(module, br, context);
+				CustomModifiers mods = CustomModifiers.Read(module, br, context);
 				// C++/CLI generates void& local variables, so we need to use ReadTypeOrVoid here
-				return ReadTypeOrVoid(module, br, context).__MakeByRefType(mods.required, mods.optional);
+				return ReadTypeOrVoid(module, br, context).__MakeByRefType(mods);
 			}
 			else
 			{
@@ -331,18 +349,17 @@ namespace IKVM.Reflection
 					int rank = type.GetArrayRank();
 					bb.Write(ELEMENT_TYPE_ARRAY);
 					// LAMESPEC the Type production (23.2.12) doesn't include CustomMod* for arrays, but the verifier allows it and ildasm also supports it
-					WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_REQD, type.__GetRequiredCustomModifiers());
-					WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_OPT, type.__GetOptionalCustomModifiers());
+					WriteCustomModifiers(module, bb, type.__GetCustomModifiers());
 					WriteType(module, bb, type.GetElementType());
-					bb.WriteCompressedInt(rank);
+					bb.WriteCompressedUInt(rank);
 					int[] sizes = type.__GetArraySizes();
-					bb.WriteCompressedInt(sizes.Length);
+					bb.WriteCompressedUInt(sizes.Length);
 					for (int i = 0; i < sizes.Length; i++)
 					{
-						bb.WriteCompressedInt(sizes[i]);
+						bb.WriteCompressedUInt(sizes[i]);
 					}
 					int[] lobounds = type.__GetArrayLowerBounds();
-					bb.WriteCompressedInt(lobounds.Length);
+					bb.WriteCompressedUInt(lobounds.Length);
 					for (int i = 0; i < lobounds.Length; i++)
 					{
 						bb.WriteCompressedInt(lobounds[i]);
@@ -357,8 +374,7 @@ namespace IKVM.Reflection
 				{
 					bb.Write(ELEMENT_TYPE_PTR);
 				}
-				WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_REQD, type.__GetRequiredCustomModifiers());
-				WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_OPT, type.__GetOptionalCustomModifiers());
+				WriteCustomModifiers(module, bb, type.__GetCustomModifiers());
 				type = type.GetElementType();
 			}
 			Universe u = module.universe;
@@ -444,12 +460,17 @@ namespace IKVM.Reflection
 				{
 					bb.Write(ELEMENT_TYPE_VAR);
 				}
-				bb.WriteCompressedInt(type.GenericParameterPosition);
+				bb.WriteCompressedUInt(type.GenericParameterPosition);
 			}
 			else if (!type.__IsMissing && type.IsGenericType)
 			{
 				WriteGenericSignature(module, bb, type);
 			}
+			else if (type.__IsFunctionPointer)
+			{
+				bb.Write(ELEMENT_TYPE_FNPTR);
+				WriteStandAloneMethodSig(module, bb, type.__MethodSignature);
+			}
 			else
 			{
 				if (type.IsValueType)
@@ -467,8 +488,7 @@ namespace IKVM.Reflection
 		private static void WriteGenericSignature(ModuleBuilder module, ByteBuffer bb, Type type)
 		{
 			Type[] typeArguments = type.GetGenericArguments();
-			Type[][] requiredCustomModifiers = type.__GetGenericArgumentsRequiredCustomModifiers();
-			Type[][] optionalCustomModifiers = type.__GetGenericArgumentsOptionalCustomModifiers();
+			CustomModifiers[] customModifiers = type.__GetGenericArgumentsCustomModifiers();
 			if (!type.IsGenericTypeDefinition)
 			{
 				type = type.GetGenericTypeDefinition();
@@ -483,73 +503,26 @@ namespace IKVM.Reflection
 				bb.Write(ELEMENT_TYPE_CLASS);
 			}
 			bb.WriteTypeDefOrRefEncoded(module.GetTypeToken(type).Token);
-			bb.WriteCompressedInt(typeArguments.Length);
+			bb.WriteCompressedUInt(typeArguments.Length);
 			for (int i = 0; i < typeArguments.Length; i++)
 			{
-				WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_REQD, requiredCustomModifiers[i]);
-				WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_OPT, optionalCustomModifiers[i]);
+				WriteCustomModifiers(module, bb, customModifiers[i]);
 				WriteType(module, bb, typeArguments[i]);
 			}
 		}
 
-		protected static void WriteCustomModifiers(ModuleBuilder module, ByteBuffer bb, byte mod, Type[] modifiers)
+		protected static void WriteCustomModifiers(ModuleBuilder module, ByteBuffer bb, CustomModifiers modifiers)
 		{
-			if (modifiers != null)
+			foreach (CustomModifiers.Entry entry in modifiers)
 			{
-				foreach (Type type in modifiers)
-				{
-					bb.Write(mod);
-					bb.WriteTypeDefOrRefEncoded(module.GetTypeTokenForMemberRef(type));
-				}
+				bb.Write(entry.IsRequired ? ELEMENT_TYPE_CMOD_REQD : ELEMENT_TYPE_CMOD_OPT);
+				bb.WriteTypeDefOrRefEncoded(module.GetTypeTokenForMemberRef(entry.Type));
 			}
 		}
 
-		protected static bool IsCustomModifier(byte b)
+		internal static Type ReadTypeDefOrRefEncoded(ModuleReader module, ByteReader br, IGenericContext context)
 		{
-			return b == ELEMENT_TYPE_CMOD_OPT || b == ELEMENT_TYPE_CMOD_REQD;
-		}
-
-		struct CustomModifiers
-		{
-			internal Type[] required;
-			internal Type[] optional;
-		}
-
-		private static CustomModifiers ReadCustomModifiers(ModuleReader module, ByteReader br, IGenericContext context)
-		{
-			CustomModifiers mods = new CustomModifiers();
-			byte b = br.PeekByte();
-			if (IsCustomModifier(b))
-			{
-				List<Type> required = new List<Type>();
-				List<Type> optional = new List<Type>();
-				while (IsCustomModifier(b))
-				{
-					bool req = br.ReadByte() == ELEMENT_TYPE_CMOD_REQD;
-					Type type = ReadTypeDefOrRefEncoded(module, br, context);
-					(req ? required : optional).Add(type);
-					b = br.PeekByte();
-				}
-				mods.required = required.ToArray();
-				mods.optional = optional.ToArray();
-			}
-			return mods;
-		}
-
-		protected static void SkipCustomModifiers(ByteReader br)
-		{
-			byte b = br.PeekByte();
-			while (IsCustomModifier(b))
-			{
-				br.ReadByte();
-				br.ReadCompressedInt();
-				b = br.PeekByte();
-			}
-		}
-
-		private static Type ReadTypeDefOrRefEncoded(ModuleReader module, ByteReader br, IGenericContext context)
-		{
-			int encoded = br.ReadCompressedInt();
+			int encoded = br.ReadCompressedUInt();
 			switch (encoded & 3)
 			{
 				case 0:
@@ -563,152 +536,70 @@ namespace IKVM.Reflection
 			}
 		}
 
-		protected static void ReadCustomModifiers(ModuleReader module, ByteReader br, IGenericContext context, out Type[] requiredCustomModifiers, out Type[] optionalCustomModifiers)
+		internal static void WriteStandAloneMethodSig(ModuleBuilder module, ByteBuffer bb, __StandAloneMethodSig sig)
 		{
-			byte b = br.PeekByte();
-			if (IsCustomModifier(b))
+			if (sig.IsUnmanaged)
 			{
-				List<Type> required = new List<Type>();
-				List<Type> optional = new List<Type>();
-				while (IsCustomModifier(b))
+				switch (sig.UnmanagedCallingConvention)
 				{
-					br.ReadByte();
-					Type type = ReadTypeDefOrRefEncoded(module, br, context);
-					if (b == ELEMENT_TYPE_CMOD_REQD)
-					{
-						required.Add(type);
-					}
-					else
-					{
-						optional.Add(type);
-					}
-					b = br.PeekByte();
+					case CallingConvention.Cdecl:
+						bb.Write((byte)0x01);	// C
+						break;
+					case CallingConvention.StdCall:
+					case CallingConvention.Winapi:
+						bb.Write((byte)0x02);	// STDCALL
+						break;
+					case CallingConvention.ThisCall:
+						bb.Write((byte)0x03);	// THISCALL
+						break;
+					case CallingConvention.FastCall:
+						bb.Write((byte)0x04);	// FASTCALL
+						break;
+					default:
+						throw new ArgumentOutOfRangeException("callingConvention");
 				}
-				requiredCustomModifiers = required.ToArray();
-				optionalCustomModifiers = optional.ToArray();
 			}
 			else
 			{
-				requiredCustomModifiers = null;
-				optionalCustomModifiers = null;
+				CallingConventions callingConvention = sig.CallingConvention;
+				byte flags = 0;
+				if ((callingConvention & CallingConventions.HasThis) != 0)
+				{
+					flags |= HASTHIS;
+				}
+				if ((callingConvention & CallingConventions.ExplicitThis) != 0)
+				{
+					flags |= EXPLICITTHIS;
+				}
+				if ((callingConvention & CallingConventions.VarArgs) != 0)
+				{
+					flags |= VARARG;
+				}
+				bb.Write(flags);
 			}
-		}
-
-		// unmanaged calling convention
-		internal static void WriteStandAloneMethodSig(ModuleBuilder module, ByteBuffer bb, CallingConvention callingConvention, Type returnType, Type[] parameterTypes)
-		{
-			switch (callingConvention)
-			{
-				case CallingConvention.Cdecl:
-					bb.Write((byte)0x01);	// C
-					break;
-				case CallingConvention.StdCall:
-				case CallingConvention.Winapi:
-					bb.Write((byte)0x02);	// STDCALL
-					break;
-				case CallingConvention.ThisCall:
-					bb.Write((byte)0x03);	// THISCALL
-					break;
-				case CallingConvention.FastCall:
-					bb.Write((byte)0x04);	// FASTCALL
-					break;
-				default:
-					throw new ArgumentOutOfRangeException("callingConvention");
-			}
-			bb.WriteCompressedInt(parameterTypes.Length);
-			WriteType(module, bb, returnType);
-			foreach (Type t in parameterTypes)
-			{
-				WriteType(module, bb, t);
-			}
-		}
-
-		// managed calling convention
-		internal static void WriteStandAloneMethodSig(ModuleBuilder module, ByteBuffer bb, CallingConventions callingConvention, Type returnType, Type[] parameterTypes, Type[] optionalParameterTypes)
-		{
-			byte flags = 0;
-			if ((callingConvention & CallingConventions.HasThis) != 0)
-			{
-				flags |= HASTHIS;
-			}
-			if ((callingConvention & CallingConventions.ExplicitThis) != 0)
-			{
-				flags |= EXPLICITTHIS;
-			}
-			if ((callingConvention & CallingConventions.VarArgs) != 0)
-			{
-				flags |= VARARG;
-			}
-			bb.Write(flags);
-			bb.WriteCompressedInt(parameterTypes.Length + optionalParameterTypes.Length);
-			WriteType(module, bb, returnType);
+			Type[] parameterTypes = sig.ParameterTypes;
+			Type[] optionalParameterTypes = sig.OptionalParameterTypes;
+			bb.WriteCompressedUInt(parameterTypes.Length + optionalParameterTypes.Length);
+			WriteCustomModifiers(module, bb, sig.GetReturnTypeCustomModifiers());
+			WriteType(module, bb, sig.ReturnType);
+			int index = 0;
 			foreach (Type t in parameterTypes)
 			{
+				WriteCustomModifiers(module, bb, sig.GetParameterCustomModifiers(index++));
 				WriteType(module, bb, t);
 			}
+			// note that optional parameters are only allowed for managed signatures (but we don't enforce that)
 			if (optionalParameterTypes.Length > 0)
 			{
 				bb.Write(SENTINEL);
 				foreach (Type t in optionalParameterTypes)
 				{
+					WriteCustomModifiers(module, bb, sig.GetParameterCustomModifiers(index++));
 					WriteType(module, bb, t);
 				}
 			}
 		}
 
-		internal static void WriteLocalVarSig(ModuleBuilder module, ByteBuffer bb, IList<LocalBuilder> locals)
-		{
-			bb.Write(LOCAL_SIG);
-			bb.WriteCompressedInt(locals.Count);
-			foreach (LocalBuilder local in locals)
-			{
-				if (local.IsPinned)
-				{
-					bb.Write(ELEMENT_TYPE_PINNED);
-				}
-				WriteType(module, bb, local.LocalType);
-			}
-		}
-
-		internal static void WritePropertySig(ModuleBuilder module, ByteBuffer bb, CallingConventions callingConvention,
-			Type returnType, Type[] returnTypeRequiredCustomModifiers, Type[] returnTypeOptionalCustomModifiers,
-			Type[] parameterTypes, Type[][] parameterTypeRequiredCustomModifiers, Type[][] parameterTypeOptionalCustomModifiers)
-		{
-			byte flags = PROPERTY;
-			if ((callingConvention & CallingConventions.HasThis) != 0)
-			{
-				flags |= HASTHIS;
-			}
-			if ((callingConvention & CallingConventions.ExplicitThis) != 0)
-			{
-				flags |= EXPLICITTHIS;
-			}
-			if ((callingConvention & CallingConventions.VarArgs) != 0)
-			{
-				flags |= VARARG;
-			}
-			bb.Write(flags);
-			bb.WriteCompressedInt(parameterTypes == null ? 0 : parameterTypes.Length);
-			WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_REQD, returnTypeRequiredCustomModifiers);
-			WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_OPT, returnTypeOptionalCustomModifiers);
-			WriteType(module, bb, returnType);
-			if (parameterTypes != null)
-			{
-				for (int i = 0; i < parameterTypes.Length; i++)
-				{
-					if (parameterTypeRequiredCustomModifiers != null)
-					{
-						WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_REQD, parameterTypeRequiredCustomModifiers[i]);
-					}
-					if (parameterTypeOptionalCustomModifiers != null)
-					{
-						WriteCustomModifiers(module, bb, ELEMENT_TYPE_CMOD_OPT, parameterTypeOptionalCustomModifiers[i]);
-					}
-					WriteType(module, bb, parameterTypes[i]);
-				}
-			}
-		}
-
 		internal static void WriteTypeSpec(ModuleBuilder module, ByteBuffer bb, Type type)
 		{
 			WriteType(module, bb, type);
@@ -717,7 +608,7 @@ namespace IKVM.Reflection
 		internal static void WriteMethodSpec(ModuleBuilder module, ByteBuffer bb, Type[] genArgs)
 		{
 			bb.Write(GENERICINST);
-			bb.WriteCompressedInt(genArgs.Length);
+			bb.WriteCompressedUInt(genArgs.Length);
 			foreach (Type arg in genArgs)
 			{
 				WriteType(module, bb, arg);
@@ -725,28 +616,30 @@ namespace IKVM.Reflection
 		}
 
 		// this reads just the optional parameter types, from a MethodRefSig
-		internal static Type[] ReadOptionalParameterTypes(ModuleReader module, ByteReader br)
+		internal static Type[] ReadOptionalParameterTypes(ModuleReader module, ByteReader br, IGenericContext context, out CustomModifiers[] customModifiers)
 		{
 			br.ReadByte();
-			int paramCount = br.ReadCompressedInt();
-			SkipCustomModifiers(br);
-			ReadRetType(module, br, null);
+			int paramCount = br.ReadCompressedUInt();
+			CustomModifiers.Skip(br);
+			ReadRetType(module, br, context);
 			for (int i = 0; i < paramCount; i++)
 			{
 				if (br.PeekByte() == SENTINEL)
 				{
 					br.ReadByte();
 					Type[] types = new Type[paramCount - i];
+					customModifiers = new CustomModifiers[types.Length];
 					for (int j = 0; j < types.Length; j++)
 					{
-						SkipCustomModifiers(br);
-						types[j] = ReadType(module, br, null);
+						customModifiers[j] = CustomModifiers.Read(module, br, context);
+						types[j] = ReadType(module, br, context);
 					}
 					return types;
 				}
-				SkipCustomModifiers(br);
-				ReadType(module, br, null);
+				CustomModifiers.Skip(br);
+				ReadType(module, br, context);
 			}
+			customModifiers = Empty<CustomModifiers>.Array;
 			return Type.EmptyTypes;
 		}
 
@@ -764,41 +657,40 @@ namespace IKVM.Reflection
 			return expanded;
 		}
 
-		protected static Type[][] BindTypeParameters(IGenericBinder binder, Type[][] types)
+		internal static void WriteSignatureHelper(ModuleBuilder module, ByteBuffer bb, byte flags, ushort paramCount, List<Type> args)
 		{
-			if (types == null)
+			bb.Write(flags);
+			if (flags != FIELD)
 			{
-				return null;
+				bb.WriteCompressedUInt(paramCount);
 			}
-			Type[][] expanded = new Type[types.Length][];
-			for (int i = 0; i < types.Length; i++)
+			foreach (Type type in args)
 			{
-				expanded[i] = BindTypeParameters(binder, types[i]);
+				if (type == MarkerType.ModOpt)
+				{
+					bb.Write(ELEMENT_TYPE_CMOD_OPT);
+				}
+				else if (type == MarkerType.ModReq)
+				{
+					bb.Write(ELEMENT_TYPE_CMOD_REQD);
+				}
+				else if (type == MarkerType.Sentinel)
+				{
+					bb.Write(SENTINEL);
+				}
+				else if (type == MarkerType.Pinned)
+				{
+					bb.Write(ELEMENT_TYPE_PINNED);
+				}
+				else if (type == null)
+				{
+					bb.Write(ELEMENT_TYPE_VOID);
+				}
+				else
+				{
+					WriteType(module, bb, type);
+				}
 			}
-			return expanded;
-		}
-
-		protected static Type[][][] BindTypeParameters(IGenericBinder binder, Type[][][] types)
-		{
-			if (types == null)
-			{
-				return null;
-			}
-			Type[][][] expanded = new Type[types.Length][][];
-			for (int i = 0; i < types.Length; i++)
-			{
-				expanded[i] = BindTypeParameters(binder, types[i]);
-			}
-			return expanded;
-		}
-
-		protected static Type[] BindTypeParameters(IGenericBinder binder, Type[][][] types, int index, int optOrReq)
-		{
-			if (types == null || types[index] == null)
-			{
-				return null;
-			}
-			return BindTypeParameters(binder, types[index][optOrReq]);
 		}
 	}
 }
