@@ -148,6 +148,9 @@ namespace IKVM.Internal
 				{
 					return assembly.GetType(name);
 				}
+				catch (ArgumentException)
+				{
+				}
 				catch (FileLoadException x)
 				{
 					// this can only happen if the assembly was loaded in the ReflectionOnly
@@ -165,6 +168,9 @@ namespace IKVM.Internal
 				{
 					return mod.GetType(name);
 				}
+				catch (ArgumentException)
+				{
+				}
 				catch (FileLoadException x)
 				{
 					// this can only happen if the assembly was loaded in the ReflectionOnly
@@ -246,6 +252,7 @@ namespace IKVM.Internal
 				{
 					// for fake types, we load the declaring outer type (the real one) and
 					// let that generated the manufactured nested classes
+					// (note that for generic outer types, we need to duplicate this in ClassLoaderWrapper.LoadGenericClass)
 					TypeWrapper outer = null;
 					if (name.EndsWith(DotNetTypeWrapper.DelegateInterfaceSuffix))
 					{
@@ -344,7 +351,9 @@ namespace IKVM.Internal
 				}
 				foreach (AssemblyName name in internalsVisibleTo)
 				{
-					if (AssemblyName.ReferenceMatchesDefinition(name, otherName))
+					// we match the simple name and PublicKeyToken (because the AssemblyName constructor used
+					// by GetInternalsVisibleToAttributes() only sets the PublicKeyToken, even if a PublicKey is specified)
+					if (ReflectUtil.MatchNameAndPublicKeyToken(name, otherName))
 					{
 						return true;
 					}
@@ -674,8 +683,12 @@ namespace IKVM.Internal
 				{
 					// this really shouldn't happen, it means that we have two different types in our assembly that both
 					// have the same Java name
+#if STATIC_COMPILER
+					throw new FatalCompilerErrorException(Message.AssemblyContainsDuplicateClassNames, type.FullName, wrapper.TypeAsTBD.FullName, wrapper.Name, type.Assembly.FullName);
+#else
 					string msg = String.Format("\nType \"{0}\" and \"{1}\" both map to the same name \"{2}\".\n", type.FullName, wrapper.TypeAsTBD.FullName, wrapper.Name);
 					JVM.CriticalFailure(msg, null);
+#endif
 				}
 				return wrapper;
 			}
@@ -937,7 +950,7 @@ namespace IKVM.Internal
 			}
 			otherName = acl.GetAssembly(friend).GetName();
 #endif
-			return GetLoaderForExportedAssembly(GetAssembly(wrapper)).InternalsVisibleTo(otherName);
+			return GetLoader(GetAssembly(wrapper)).InternalsVisibleTo(otherName);
 		}
 
 		// this method should not be used with dynamic Java assemblies
@@ -1017,6 +1030,7 @@ namespace IKVM.Internal
 				foreach (KeyValuePair<string, string> kv in customClassLoaderRedirects)
 				{
 					string asm = kv.Key;
+					// FXBUG
 					// We only support matching on the assembly's simple name,
 					// because there appears to be no viable alternative.
 					// There is AssemblyName.ReferenceMatchesDefinition()
