@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2011 Jeroen Frijters
+  Copyright (C) 2002-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -51,12 +51,12 @@ namespace IKVM.Internal
 		None = 0,
 		LocalVariableTable = 1,
 		LineNumberTable = 2,
+		RelaxedClassNameValidation = 4,
 	}
 
 	sealed class ClassFile
 	{
 		private ConstantPoolItem[] constantpool;
-		private string[] utf8_cp;
 		// Modifiers is a ushort, so the next four fields combine into two 32 bit slots
 		private Modifiers access_flags;
 		private ushort this_class;
@@ -71,6 +71,9 @@ namespace IKVM.Internal
 		private Field[] fields;
 		private Method[] methods;
 		private string sourceFile;
+#if STATIC_COMPILER
+		private string sourcePath;
+#endif
 		private string ikvmAssembly;
 		private InnerClass[] innerClasses;
 		private object[] annotations;
@@ -178,7 +181,7 @@ namespace IKVM.Internal
 				flags = majorVersion;
 				int constantpoolcount = br.ReadUInt16();
 				constantpool = new ConstantPoolItem[constantpoolcount];
-				utf8_cp = new string[constantpoolcount];
+				string[] utf8_cp = new string[constantpoolcount];
 				for(int i = 1; i < constantpoolcount; i++)
 				{
 					Constant tag = (Constant)br.ReadByte();
@@ -244,7 +247,7 @@ namespace IKVM.Internal
 					{
 						try
 						{
-							constantpool[i].Resolve(this);
+							constantpool[i].Resolve(this, utf8_cp, options);
 						}
 						catch(ClassFormatError x)
 						{
@@ -316,7 +319,7 @@ namespace IKVM.Internal
 				fields = new Field[fields_count];
 				for(int i = 0; i < fields_count; i++)
 				{
-					fields[i] = new Field(this, br);
+					fields[i] = new Field(this, utf8_cp, br);
 					string name = fields[i].Name;
 					if(!IsValidFieldName(name, majorVersion))
 					{
@@ -334,7 +337,7 @@ namespace IKVM.Internal
 				methods = new Method[methods_count];
 				for(int i = 0; i < methods_count; i++)
 				{
-					methods[i] = new Method(this, options, br);
+					methods[i] = new Method(this, utf8_cp, options, br);
 					string name = methods[i].Name;
 					string sig = methods[i].Signature;
 					if(!IsValidMethodName(name, majorVersion))
@@ -359,7 +362,7 @@ namespace IKVM.Internal
 				int attributes_count = br.ReadUInt16();
 				for(int i = 0; i < attributes_count; i++)
 				{
-					switch(GetConstantPoolUtf8String(br.ReadUInt16()))
+					switch(GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16()))
 					{
 						case "Deprecated":
 							if(br.ReadUInt32() != 0)
@@ -373,7 +376,7 @@ namespace IKVM.Internal
 							{
 								throw new ClassFormatError("SourceFile attribute has incorrect length");
 							}
-							sourceFile = GetConstantPoolUtf8String(br.ReadUInt16());
+							sourceFile = GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 							break;
 						case "InnerClasses":
 						{
@@ -407,6 +410,11 @@ namespace IKVM.Internal
 								{
 									throw new ClassFormatError("{0} (Class is both inner and outer class)", this.Name);
 								}
+								if(innerClasses[j].innerClass != 0 && innerClasses[j].outerClass != 0)
+								{
+									MarkLinkRequiredConstantPoolItem(innerClasses[j].innerClass);
+									MarkLinkRequiredConstantPoolItem(innerClasses[j].outerClass);
+								}
 							}
 							break;
 						}
@@ -419,7 +427,7 @@ namespace IKVM.Internal
 							{
 								throw new ClassFormatError("Signature attribute has incorrect length");
 							}
-							signature = GetConstantPoolUtf8String(br.ReadUInt16());
+							signature = GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 							break;
 						case "EnclosingMethod":
 							if(majorVersion < 49)
@@ -447,8 +455,8 @@ namespace IKVM.Internal
 									ConstantPoolItemNameAndType m = (ConstantPoolItemNameAndType)GetConstantPoolItem(method_index);
 									enclosingMethod = new string[] {
 										GetConstantPoolClass(class_index),
-										GetConstantPoolUtf8String(m.name_index),
-										GetConstantPoolUtf8String(m.descriptor_index).Replace('/', '.')
+										GetConstantPoolUtf8String(utf8_cp, m.name_index),
+										GetConstantPoolUtf8String(utf8_cp, m.descriptor_index).Replace('/', '.')
 																   };
 								}
 							}
@@ -458,7 +466,7 @@ namespace IKVM.Internal
 							{
 								goto default;
 							}
-							annotations = ReadAnnotations(br, this);
+							annotations = ReadAnnotations(br, this, utf8_cp);
 							break;
 #if STATIC_COMPILER
 						case "RuntimeInvisibleAnnotations":
@@ -466,7 +474,7 @@ namespace IKVM.Internal
 							{
 								goto default;
 							}
-							foreach(object[] annot in ReadAnnotations(br, this))
+							foreach(object[] annot in ReadAnnotations(br, this, utf8_cp))
 							{
 								if(annot[1].Equals("Likvm/lang/Internal;"))
 								{
@@ -488,7 +496,7 @@ namespace IKVM.Internal
 							{
 								throw new ClassFormatError("IKVM.NET.Assembly attribute has incorrect length");
 							}
-							ikvmAssembly = GetConstantPoolUtf8String(br.ReadUInt16());
+							ikvmAssembly = GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 							break;
 						default:
 							br.Skip(br.ReadUInt32());
@@ -508,9 +516,6 @@ namespace IKVM.Internal
 						}
 					}
 				}
-				// now that we've constructed the high level objects, the utf8 table isn't needed anymore
-				// TODO remove utf8_cp field from ClassFile object
-				utf8_cp = null;
 				if(br.Position != offset + length)
 				{
 					throw new ClassFormatError("Extra bytes at the end of the class file");
@@ -535,6 +540,14 @@ namespace IKVM.Internal
 			//		}
 		}
 
+		private void MarkLinkRequiredConstantPoolItem(int index)
+		{
+			if (index > 0 && index < constantpool.Length && constantpool[index] != null)
+			{
+				constantpool[index].MarkLinkRequired();
+			}
+		}
+
 		private static BootstrapMethod[] ReadBootstrapMethods(BigEndianBinaryReader br, ClassFile classFile)
 		{
 			BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
@@ -547,6 +560,7 @@ namespace IKVM.Internal
 				{
 					throw new ClassFormatError("bootstrap_method_index {0} has bad constant type in class file {1}", bsm_index, classFile.Name);
 				}
+				classFile.MarkLinkRequiredConstantPoolItem(bsm_index);
 				ushort argument_count = rdr.ReadUInt16();
 				ushort[] args = new ushort[argument_count];
 				for(int j = 0; j < args.Length; j++)
@@ -556,6 +570,7 @@ namespace IKVM.Internal
 					{
 						throw new ClassFormatError("argument_index {0} has bad constant type in class file {1}", argument_index, classFile.Name);
 					}
+					classFile.MarkLinkRequiredConstantPoolItem(argument_index);
 					args[j] = argument_index;
 				}
 				bsm[i] = new BootstrapMethod(bsm_index, args);
@@ -581,14 +596,14 @@ namespace IKVM.Internal
 			return false;
 		}
 
-		private static object[] ReadAnnotations(BigEndianBinaryReader br, ClassFile classFile)
+		private static object[] ReadAnnotations(BigEndianBinaryReader br, ClassFile classFile, string[] utf8_cp)
 		{
 			BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
 			ushort num_annotations = rdr.ReadUInt16();
 			object[] annotations = new object[num_annotations];
 			for(int i = 0; i < annotations.Length; i++)
 			{
-				annotations[i] = ReadAnnotation(rdr, classFile);
+				annotations[i] = ReadAnnotation(rdr, classFile, utf8_cp);
 			}
 			if(!rdr.IsAtEnd)
 			{
@@ -597,22 +612,22 @@ namespace IKVM.Internal
 			return annotations;
 		}
 
-		private static object ReadAnnotation(BigEndianBinaryReader rdr, ClassFile classFile)
+		private static object ReadAnnotation(BigEndianBinaryReader rdr, ClassFile classFile, string[] utf8_cp)
 		{
-			string type = classFile.GetConstantPoolUtf8String(rdr.ReadUInt16());
+			string type = classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16());
 			ushort num_element_value_pairs = rdr.ReadUInt16();
 			object[] annot = new object[2 + num_element_value_pairs * 2];
 			annot[0] = AnnotationDefaultAttribute.TAG_ANNOTATION;
 			annot[1] = type;
 			for(int i = 0; i < num_element_value_pairs; i++)
 			{
-				annot[2 + i * 2 + 0] = classFile.GetConstantPoolUtf8String(rdr.ReadUInt16());
-				annot[2 + i * 2 + 1] = ReadAnnotationElementValue(rdr, classFile);
+				annot[2 + i * 2 + 0] = classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16());
+				annot[2 + i * 2 + 1] = ReadAnnotationElementValue(rdr, classFile, utf8_cp);
 			}
 			return annot;
 		}
 
-		private static object ReadAnnotationElementValue(BigEndianBinaryReader rdr, ClassFile classFile)
+		private static object ReadAnnotationElementValue(BigEndianBinaryReader rdr, ClassFile classFile, string[] utf8_cp)
 		{
 			try
 			{
@@ -636,24 +651,24 @@ namespace IKVM.Internal
 					case (byte)'D':
 						return classFile.GetConstantPoolConstantDouble(rdr.ReadUInt16());
 					case (byte)'s':
-						return classFile.GetConstantPoolUtf8String(rdr.ReadUInt16());
+						return classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16());
 					case (byte)'e':
 						{
 							ushort type_name_index = rdr.ReadUInt16();
 							ushort const_name_index = rdr.ReadUInt16();
 							return new object[] {
 											AnnotationDefaultAttribute.TAG_ENUM,
-											classFile.GetConstantPoolUtf8String(type_name_index),
-											classFile.GetConstantPoolUtf8String(const_name_index)
+											classFile.GetConstantPoolUtf8String(utf8_cp, type_name_index),
+											classFile.GetConstantPoolUtf8String(utf8_cp, const_name_index)
 										};
 						}
 					case (byte)'c':
 						return new object[] {
 											AnnotationDefaultAttribute.TAG_CLASS,
-											classFile.GetConstantPoolUtf8String(rdr.ReadUInt16())
+											classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16())
 										};
 					case (byte)'@':
-						return ReadAnnotation(rdr, classFile);
+						return ReadAnnotation(rdr, classFile, utf8_cp);
 					case (byte)'[':
 						{
 							ushort num_values = rdr.ReadUInt16();
@@ -661,7 +676,7 @@ namespace IKVM.Internal
 							array[0] = AnnotationDefaultAttribute.TAG_ARRAY;
 							for (int i = 0; i < num_values; i++)
 							{
-								array[i + 1] = ReadAnnotationElementValue(rdr, classFile);
+								array[i + 1] = ReadAnnotationElementValue(rdr, classFile, utf8_cp);
 							}
 							return array;
 						}
@@ -1012,7 +1027,7 @@ namespace IKVM.Internal
 			return ((ConstantPoolItemClass)constantpool[index]).GetClassType();
 		}
 
-		internal string GetConstantPoolUtf8String(int index)
+		private string GetConstantPoolUtf8String(string[] utf8_cp, int index)
 		{
 			string s = utf8_cp[index];
 			if(s == null)
@@ -1117,6 +1132,16 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal string SourcePath
+		{
+#if STATIC_COMPILER
+			get { return sourcePath; }
+			set { sourcePath = value; }
+#else
+			get { return sourceFile; }
+#endif
+		}
+
 		internal object[] Annotations
 		{
 			get
@@ -1275,7 +1300,7 @@ namespace IKVM.Internal
 
 		internal abstract class ConstantPoolItem
 		{
-			internal virtual void Resolve(ClassFile classFile)
+			internal virtual void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
 			}
 
@@ -1287,6 +1312,10 @@ namespace IKVM.Internal
 			{
 				throw new InvalidOperationException();
 			}
+
+			internal virtual void MarkLinkRequired()
+			{
+			}
 		}
 
 		internal sealed class ConstantPoolItemClass : ConstantPoolItem
@@ -1301,12 +1330,16 @@ namespace IKVM.Internal
 				name_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
-				name = classFile.GetConstantPoolUtf8String(name_index);
+				name = classFile.GetConstantPoolUtf8String(utf8_cp, name_index);
 				if(name.Length > 0)
 				{
-					if(classFile.MajorVersion < 49)
+					// We don't enforce the strict class name rules in the static compiler, since HotSpot doesn't enforce *any* rules on
+					// class names for the system (and boot) class loader. We still need to enforce the 1.5 restrictions, because we
+					// rely on those invariants.
+#if !STATIC_COMPILER
+					if(classFile.MajorVersion < 49 && (options & ClassFileParseOptions.RelaxedClassNameValidation) == 0)
 					{
 						char prev = name[0];
 						if(Char.IsLetter(prev) || prev == '$' || prev == '_' || prev == '[' || prev == '/')
@@ -1342,6 +1375,7 @@ namespace IKVM.Internal
 						}
 					}
 					else
+#endif
 					{
 						// since 1.5 the restrictions on class names have been greatly reduced
 						int end = name.Length;
@@ -1372,9 +1406,14 @@ namespace IKVM.Internal
 					throw new ClassFormatError("Invalid class name \"{0}\"", name);
 			}
 
+			internal override void MarkLinkRequired()
+			{
+				typeWrapper = VerifierTypeWrapper.Null;
+			}
+
 			internal override void Link(TypeWrapper thisType)
 			{
-				if(typeWrapper == null)
+				if(typeWrapper == VerifierTypeWrapper.Null)
 				{
 					typeWrapper = ClassLoaderWrapper.LoadClassNoThrow(thisType.GetClassLoader(), name);
 				}
@@ -1436,7 +1475,7 @@ namespace IKVM.Internal
 				name_and_type_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
 				ConstantPoolItemNameAndType name_and_type = (ConstantPoolItemNameAndType)classFile.GetConstantPoolItem(name_and_type_index);
 				clazz = (ConstantPoolItemClass)classFile.GetConstantPoolItem(class_index);
@@ -1445,14 +1484,19 @@ namespace IKVM.Internal
 				{
 					throw new ClassFormatError("Bad index in constant pool");
 				}
-				name = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.name_index));
-				descriptor = classFile.GetConstantPoolUtf8String(name_and_type.descriptor_index);
+				name = String.Intern(classFile.GetConstantPoolUtf8String(utf8_cp, name_and_type.name_index));
+				descriptor = classFile.GetConstantPoolUtf8String(utf8_cp, name_and_type.descriptor_index);
 				Validate(name, descriptor, classFile.MajorVersion);
 				descriptor = String.Intern(descriptor.Replace('/', '.'));
 			}
 
 			protected abstract void Validate(string name, string descriptor, int majorVersion);
 
+			internal override void MarkLinkRequired()
+			{
+				clazz.MarkLinkRequired();
+			}
+
 			internal override void Link(TypeWrapper thisType)
 			{
 				clazz.Link(thisType);
@@ -1528,6 +1572,10 @@ namespace IKVM.Internal
 				}
 				FieldWrapper fw = null;
 				TypeWrapper wrapper = GetClassType();
+				if(wrapper == null)
+				{
+					return;
+				}
 				if(!wrapper.IsUnloadable)
 				{
 					fw = wrapper.GetFieldWrapper(Name, Signature);
@@ -1648,7 +1696,7 @@ namespace IKVM.Internal
 			{
 				base.Link(thisType);
 				TypeWrapper wrapper = GetClassType();
-				if(!wrapper.IsUnloadable)
+				if(wrapper != null && !wrapper.IsUnloadable)
 				{
 					method = wrapper.GetMethodWrapper(Name, Signature, !ReferenceEquals(Name, StringConstants.INIT));
 					if(method != null)
@@ -1698,7 +1746,7 @@ namespace IKVM.Internal
 			{
 				base.Link(thisType);
 				TypeWrapper wrapper = GetClassType();
-				if(!wrapper.IsUnloadable)
+				if(wrapper != null && !wrapper.IsUnloadable)
 				{
 					method = GetInterfaceMethod(wrapper, Name, Signature);
 					if(method == null)
@@ -1794,10 +1842,10 @@ namespace IKVM.Internal
 				descriptor_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
-				if(classFile.GetConstantPoolUtf8String(name_index) == null
-					|| classFile.GetConstantPoolUtf8String(descriptor_index) == null)
+				if(classFile.GetConstantPoolUtf8String(utf8_cp, name_index) == null
+					|| classFile.GetConstantPoolUtf8String(utf8_cp, descriptor_index) == null)
 				{
 					throw new ClassFormatError("Illegal constant pool index");
 				}
@@ -1816,7 +1864,7 @@ namespace IKVM.Internal
 				method_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
 				switch ((RefKind)ref_kind)
 				{
@@ -1846,6 +1894,11 @@ namespace IKVM.Internal
 				}
 			}
 
+			internal override void MarkLinkRequired()
+			{
+				cpi.MarkLinkRequired();
+			}
+
 			internal string Class
 			{
 				get { return cpi.Class; }
@@ -1904,9 +1957,9 @@ namespace IKVM.Internal
 				signature_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
-				string descriptor = classFile.GetConstantPoolUtf8String(signature_index);
+				string descriptor = classFile.GetConstantPoolUtf8String(utf8_cp, signature_index);
 				if (descriptor == null || !IsValidMethodSig(descriptor))
 				{
 					throw new ClassFormatError("Invalid MethodType signature");
@@ -1967,7 +2020,7 @@ namespace IKVM.Internal
 				name_and_type_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
 				ConstantPoolItemNameAndType name_and_type = (ConstantPoolItemNameAndType)classFile.GetConstantPoolItem(name_and_type_index);
 				// if the constant pool items referred to were strings, GetConstantPoolItem returns null
@@ -1975,8 +2028,8 @@ namespace IKVM.Internal
 				{
 					throw new ClassFormatError("Bad index in constant pool");
 				}
-				name = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.name_index));
-				descriptor = String.Intern(classFile.GetConstantPoolUtf8String(name_and_type.descriptor_index).Replace('/', '.'));
+				name = String.Intern(classFile.GetConstantPoolUtf8String(utf8_cp, name_and_type.name_index));
+				descriptor = String.Intern(classFile.GetConstantPoolUtf8String(utf8_cp, name_and_type.descriptor_index).Replace('/', '.'));
 			}
 
 			internal override void Link(TypeWrapper thisType)
@@ -2032,9 +2085,9 @@ namespace IKVM.Internal
 				string_index = br.ReadUInt16();
 			}
 
-			internal override void Resolve(ClassFile classFile)
+			internal override void Resolve(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options)
 			{
-				s = classFile.GetConstantPoolUtf8String(string_index);
+				s = classFile.GetConstantPoolUtf8String(utf8_cp, string_index);
 			}
 
 			internal override ConstantType GetConstantType()
@@ -2079,11 +2132,11 @@ namespace IKVM.Internal
 			protected string signature;
 			protected object[] annotations;
 
-			internal FieldOrMethod(ClassFile classFile, BigEndianBinaryReader br)
+			internal FieldOrMethod(ClassFile classFile, string[] utf8_cp, BigEndianBinaryReader br)
 			{
 				access_flags = (Modifiers)br.ReadUInt16();
-				name = String.Intern(classFile.GetConstantPoolUtf8String(br.ReadUInt16()));
-				descriptor = classFile.GetConstantPoolUtf8String(br.ReadUInt16());
+				name = String.Intern(classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16()));
+				descriptor = classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 				ValidateSig(classFile, descriptor);
 				descriptor = String.Intern(descriptor.Replace('/', '.'));
 			}
@@ -2240,7 +2293,7 @@ namespace IKVM.Internal
 			private object constantValue;
 			private string[] propertyGetterSetter;
 
-			internal Field(ClassFile classFile, BigEndianBinaryReader br) : base(classFile, br)
+			internal Field(ClassFile classFile, string[] utf8_cp, BigEndianBinaryReader br) : base(classFile, utf8_cp, br)
 			{
 				if((IsPrivate && IsPublic) || (IsPrivate && IsProtected) || (IsPublic && IsProtected)
 					|| (IsFinal && IsVolatile)
@@ -2251,7 +2304,7 @@ namespace IKVM.Internal
 				int attributes_count = br.ReadUInt16();
 				for(int i = 0; i < attributes_count; i++)
 				{
-					switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
+					switch(classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16()))
 					{
 						case "Deprecated":
 							if(br.ReadUInt32() != 0)
@@ -2329,21 +2382,21 @@ namespace IKVM.Internal
 							{
 								throw new ClassFormatError("Signature attribute has incorrect length");
 							}
-							signature = classFile.GetConstantPoolUtf8String(br.ReadUInt16());
+							signature = classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 							break;
 						case "RuntimeVisibleAnnotations":
 							if(classFile.MajorVersion < 49)
 							{
 								goto default;
 							}
-							annotations = ReadAnnotations(br, classFile);
+							annotations = ReadAnnotations(br, classFile, utf8_cp);
 							break;
 						case "RuntimeInvisibleAnnotations":
 							if(classFile.MajorVersion < 49)
 							{
 								goto default;
 							}
-							foreach(object[] annot in ReadAnnotations(br, classFile))
+							foreach(object[] annot in ReadAnnotations(br, classFile, utf8_cp))
 							{
 								if(annot[1].Equals("Likvm/lang/Property;"))
 								{
@@ -2419,6 +2472,11 @@ namespace IKVM.Internal
 				}
 			}
 
+			internal bool IsStaticFinalConstant
+			{
+				get { return (access_flags & (Modifiers.Final | Modifiers.Static)) == (Modifiers.Final | Modifiers.Static) && constantValue != null; }
+			}
+
 			internal bool IsProperty
 			{
 				get
@@ -2448,10 +2506,19 @@ namespace IKVM.Internal
 		{
 			private Code code;
 			private string[] exceptions;
-			private object annotationDefault;
-			private object[][] parameterAnnotations;
+			private LowFreqData low;
 
-			internal Method(ClassFile classFile, ClassFileParseOptions options, BigEndianBinaryReader br) : base(classFile, br)
+			sealed class LowFreqData
+			{
+				internal object annotationDefault;
+				internal object[][] parameterAnnotations;
+#if STATIC_COMPILER
+				internal string DllExportName;
+				internal int DllExportOrdinal;
+#endif
+			}
+
+			internal Method(ClassFile classFile, string[] utf8_cp, ClassFileParseOptions options, BigEndianBinaryReader br) : base(classFile, utf8_cp, br)
 			{
 				// vmspec 4.6 says that all flags, except ACC_STRICT are ignored on <clinit>
 				// however, since Java 7 it does need to be marked static
@@ -2475,7 +2542,7 @@ namespace IKVM.Internal
 				int attributes_count = br.ReadUInt16();
 				for(int i = 0; i < attributes_count; i++)
 				{
-					switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
+					switch(classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16()))
 					{
 						case "Deprecated":
 							if(br.ReadUInt32() != 0)
@@ -2491,7 +2558,7 @@ namespace IKVM.Internal
 								throw new ClassFormatError("{0} (Duplicate Code attribute)", classFile.Name);
 							}
 							BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
-							code.Read(classFile, this, rdr, options);
+							code.Read(classFile, utf8_cp, this, rdr, options);
 							if(!rdr.IsAtEnd)
 							{
 								throw new ClassFormatError("{0} (Code attribute has wrong length)", classFile.Name);
@@ -2526,14 +2593,14 @@ namespace IKVM.Internal
 							{
 								throw new ClassFormatError("Signature attribute has incorrect length");
 							}
-							signature = classFile.GetConstantPoolUtf8String(br.ReadUInt16());
+							signature = classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16());
 							break;
 						case "RuntimeVisibleAnnotations":
 							if(classFile.MajorVersion < 49)
 							{
 								goto default;
 							}
-							annotations = ReadAnnotations(br, classFile);
+							annotations = ReadAnnotations(br, classFile, utf8_cp);
 							break;
 						case "RuntimeVisibleParameterAnnotations":
 						{
@@ -2541,16 +2608,20 @@ namespace IKVM.Internal
 							{
 								goto default;
 							}
+							if(low == null)
+							{
+								low = new LowFreqData();
+							}
 							BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
 							byte num_parameters = rdr.ReadByte();
-							parameterAnnotations = new object[num_parameters][];
+							low.parameterAnnotations = new object[num_parameters][];
 							for(int j = 0; j < num_parameters; j++)
 							{
 								ushort num_annotations = rdr.ReadUInt16();
-								parameterAnnotations[j] = new object[num_annotations];
+								low.parameterAnnotations[j] = new object[num_annotations];
 								for(int k = 0; k < num_annotations; k++)
 								{
-									parameterAnnotations[j][k] = ReadAnnotation(rdr, classFile);
+									low.parameterAnnotations[j][k] = ReadAnnotation(rdr, classFile, utf8_cp);
 								}
 							}
 							if(!rdr.IsAtEnd)
@@ -2565,8 +2636,12 @@ namespace IKVM.Internal
 							{
 								goto default;
 							}
+							if(low == null)
+							{
+								low = new LowFreqData();
+							}
 							BigEndianBinaryReader rdr = br.Section(br.ReadUInt32());
-							annotationDefault = ReadAnnotationElementValue(rdr, classFile);
+							low.annotationDefault = ReadAnnotationElementValue(rdr, classFile, utf8_cp);
 							if(!rdr.IsAtEnd)
 							{
 								throw new ClassFormatError("{0} (AnnotationDefault attribute has wrong length)", classFile.Name);
@@ -2579,7 +2654,7 @@ namespace IKVM.Internal
 							{
 								goto default;
 							}
-							foreach(object[] annot in ReadAnnotations(br, classFile))
+							foreach(object[] annot in ReadAnnotations(br, classFile, utf8_cp))
 							{
 								if(annot[1].Equals("Likvm/lang/Internal;"))
 								{
@@ -2597,6 +2672,38 @@ namespace IKVM.Internal
 								{
 									flags |= FLAG_HAS_CALLERID;
 								}
+								if(annot[1].Equals("Likvm/lang/DllExport;"))
+								{
+									string name = null;
+									int? ordinal = null;
+									for (int j = 2; j < annot.Length; j += 2)
+									{
+										if (annot[j].Equals("name") && annot[j + 1] is string)
+										{
+											name = (string)annot[j + 1];
+										}
+										else if (annot[j].Equals("ordinal") && annot[j + 1] is int)
+										{
+											ordinal = (int)annot[j + 1];
+										}
+									}
+									if (name != null && ordinal != null)
+									{
+										if (!IsStatic)
+										{
+											StaticCompiler.IssueMessage(Message.DllExportMustBeStaticMethod, classFile.Name, this.Name, this.Signature);
+										}
+										else
+										{
+											if (low == null)
+											{
+												low = new LowFreqData();
+											}
+											low.DllExportName = name;
+											low.DllExportOrdinal = ordinal.Value;
+										}
+									}
+								}
 							}
 							break;
 #endif
@@ -2651,6 +2758,14 @@ namespace IKVM.Internal
 				}
 			}
 
+			internal bool IsConstructor
+			{
+				get
+				{
+					return ReferenceEquals(Name, StringConstants.INIT);
+				}
+			}
+
 			// for use by ikvmc only
 			internal bool HasCallerIDAnnotation
 			{
@@ -2672,7 +2787,7 @@ namespace IKVM.Internal
 			{
 				get
 				{
-					return parameterAnnotations;
+					return low == null ? null : low.parameterAnnotations;
 				}
 			}
 
@@ -2680,10 +2795,28 @@ namespace IKVM.Internal
 			{
 				get
 				{
-					return annotationDefault;
+					return low == null ? null : low.annotationDefault;
 				}
 			}
 
+#if STATIC_COMPILER
+			internal string DllExportName
+			{
+				get
+				{
+					return low == null ? null : low.DllExportName;
+				}
+			}
+
+			internal int DllExportOrdinal
+			{
+				get
+				{
+					return low == null ? -1 : low.DllExportOrdinal;
+				}
+			}
+#endif
+
 			internal string VerifyError
 			{
 				get
@@ -2777,7 +2910,7 @@ namespace IKVM.Internal
 				internal LineNumberTableEntry[] lineNumberTable;
 				internal LocalVariableTableEntry[] localVariableTable;
 
-				internal void Read(ClassFile classFile, Method method, BigEndianBinaryReader br, ClassFileParseOptions options)
+				internal void Read(ClassFile classFile, string[] utf8_cp, Method method, BigEndianBinaryReader br, ClassFileParseOptions options)
 				{
 					max_stack = br.ReadUInt16();
 					max_locals = br.ReadUInt16();
@@ -2794,7 +2927,7 @@ namespace IKVM.Internal
 						BigEndianBinaryReader rdr = br.Section(code_length);
 						while(!rdr.IsAtEnd)
 						{
-							instructions[instructionIndex].Read((ushort)(rdr.Position - basePosition), rdr);
+							instructions[instructionIndex].Read((ushort)(rdr.Position - basePosition), rdr, classFile);
 							hasJsr |= instructions[instructionIndex].NormalizedOpCode == NormalizedByteCode.__jsr;
 							instructionIndex++;
 						}
@@ -2865,6 +2998,7 @@ namespace IKVM.Internal
 						{
 							throw new ClassFormatError("Illegal exception table: {0}.{1}{2}", classFile.Name, method.Name, method.Signature);
 						}
+						classFile.MarkLinkRequiredConstantPoolItem(catch_type);
 						// if start_pc, end_pc or handler_pc is invalid (i.e. doesn't point to the start of an instruction),
 						// the index will be -1 and this will be handled by the verifier
 						int startIndex = pcIndexMap[start_pc];
@@ -2886,7 +3020,7 @@ namespace IKVM.Internal
 					ushort attributes_count = br.ReadUInt16();
 					for(int i = 0; i < attributes_count; i++)
 					{
-						switch(classFile.GetConstantPoolUtf8String(br.ReadUInt16()))
+						switch(classFile.GetConstantPoolUtf8String(utf8_cp, br.ReadUInt16()))
 						{
 							case "LineNumberTable":
 								if((options & ClassFileParseOptions.LineNumberTable) != 0)
@@ -2923,8 +3057,8 @@ namespace IKVM.Internal
 									{
 										localVariableTable[j].start_pc = rdr.ReadUInt16();
 										localVariableTable[j].length = rdr.ReadUInt16();
-										localVariableTable[j].name = classFile.GetConstantPoolUtf8String(rdr.ReadUInt16());
-										localVariableTable[j].descriptor = classFile.GetConstantPoolUtf8String(rdr.ReadUInt16()).Replace('/', '.');
+										localVariableTable[j].name = classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16());
+										localVariableTable[j].descriptor = classFile.GetConstantPoolUtf8String(utf8_cp, rdr.ReadUInt16()).Replace('/', '.');
 										localVariableTable[j].index = rdr.ReadUInt16();
 									}
 									// NOTE we're intentionally not checking that we're at the end of the section
@@ -3109,7 +3243,7 @@ namespace IKVM.Internal
 					}
 				}
 
-				internal void Read(ushort pc, BigEndianBinaryReader br)
+				internal void Read(ushort pc, BigEndianBinaryReader br, ClassFile classFile)
 				{
 					this.pc = pc;
 					ByteCode bc = (ByteCode)br.ReadByte();
@@ -3118,11 +3252,15 @@ namespace IKVM.Internal
 						case ByteCodeMode.Simple:
 							break;
 						case ByteCodeMode.Constant_1:
+							arg1 = br.ReadByte();
+							classFile.MarkLinkRequiredConstantPoolItem(arg1);
+							break;
 						case ByteCodeMode.Local_1:
 							arg1 = br.ReadByte();
 							break;
 						case ByteCodeMode.Constant_2:
 							arg1 = br.ReadUInt16();
+							classFile.MarkLinkRequiredConstantPoolItem(arg1);
 							break;
 						case ByteCodeMode.Branch_2:
 							arg1 = br.ReadInt16();
@@ -3132,6 +3270,7 @@ namespace IKVM.Internal
 							break;
 						case ByteCodeMode.Constant_2_1_1:
 							arg1 = br.ReadUInt16();
+							classFile.MarkLinkRequiredConstantPoolItem(arg1);
 							arg2 = br.ReadByte();
 							if(br.ReadByte() != 0)
 							{
@@ -3150,6 +3289,7 @@ namespace IKVM.Internal
 							break;
 						case ByteCodeMode.Constant_2_Immediate_1:
 							arg1 = br.ReadUInt16();
+							classFile.MarkLinkRequiredConstantPoolItem(arg1);
 							arg2 = br.ReadSByte();
 							break;
 						case ByteCodeMode.Tableswitch:
