@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2010 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -57,7 +57,11 @@ namespace IKVM.Internal
 		internal abstract ModuleBuilder ModuleBuilder { get; }
 		internal abstract TypeWrapper DefineClassImpl(Dictionary<string, TypeWrapper> types, ClassFile f, ClassLoaderWrapper classLoader, object protectionDomain);
 		internal abstract bool ReserveName(string name);
+		internal abstract string AllocMangledName(DynamicTypeWrapper tw);
 		internal abstract Type DefineUnloadable(string name);
+#if CLASSGC
+		internal abstract void AddInternalsVisibleTo(Assembly friend);
+#endif
 	}
 #endif // !STUB_GENERATOR
 
@@ -65,9 +69,6 @@ namespace IKVM.Internal
 	{
 		private static readonly object wrapperLock = new object();
 		private static readonly Dictionary<Type, TypeWrapper> globalTypeToTypeWrapper = new Dictionary<Type, TypeWrapper>();
-#if STATIC_COMPILER
-		private TypeWrapper circularDependencyHack;
-#endif
 #if STATIC_COMPILER || STUB_GENERATOR
 		private static ClassLoaderWrapper bootstrapClassLoader;
 #else
@@ -131,7 +132,11 @@ namespace IKVM.Internal
 				}
 				else
 				{
+#if STATIC_COMPILER
+					throw new FatalCompilerErrorException(Message.CoreClassesMissing);
+#else
 					JVM.CriticalFailure("Failed to find core classes in core library", null);
+#endif
 				}
 			}
 		}
@@ -267,39 +272,19 @@ namespace IKVM.Internal
 			}
 		}
 
-		internal virtual string SourcePath
+#if !STATIC_COMPILER && !STUB_GENERATOR
+		internal bool RelaxedClassNameValidation
 		{
 			get
 			{
-				return null;
-			}
-		}
-
-#if STATIC_COMPILER
-		internal TypeWrapper LoadCircularDependencyHack(TypeWrapper tw, string name)
-		{
-			if (circularDependencyHack == null)
-			{
-				circularDependencyHack = tw;
-				try
-				{
-					return LoadClassByDottedNameFast(name);
-				}
-				finally
-				{
-					circularDependencyHack = null;
-				}
-			}
-			else if (circularDependencyHack.Name == name)
-			{
-				return circularDependencyHack;
-			}
-			else
-			{
-				return LoadClassByDottedNameFast(name);
-			}
-		}
+#if FIRST_PASS
+				return true;
+#else
+				return JVM.relaxedVerification && (javaClassLoader == null || java.lang.ClassLoader.isTrustedLoader(javaClassLoader));
 #endif
+			}
+		}
+#endif // !STATIC_COMPILER && !STUB_GENERATOR
 
 		protected virtual void CheckDefineClassAllowed(string className)
 		{
@@ -432,11 +417,6 @@ namespace IKVM.Internal
 
 		private TypeWrapper LoadClassByDottedNameFastImpl(string name, bool throwClassNotFoundException)
 		{
-			// .NET 1.1 has a limit of 1024 characters for type names
-			if(name.Length >= 1024 || name.Length == 0)
-			{
-				return null;
-			}
 			Profiler.Enter("LoadClassByDottedName");
 			try
 			{
@@ -538,6 +518,22 @@ namespace IKVM.Internal
 
 		internal TypeWrapper LoadGenericClass(string name)
 		{
+			// we need to handle delegate methods here (for generic delegates)
+			// (note that other types with manufactured inner classes such as Attribute and Enum can't be generic)
+			if (name.EndsWith(DotNetTypeWrapper.DelegateInterfaceSuffix))
+			{
+				TypeWrapper outer = LoadGenericClass(name.Substring(0, name.Length - DotNetTypeWrapper.DelegateInterfaceSuffix.Length));
+				if (outer != null && outer.IsFakeTypeContainer)
+				{
+					foreach (TypeWrapper tw in outer.InnerClasses)
+					{
+						if (tw.Name == name)
+						{
+							return tw;
+						}
+					}
+				}
+			}
 			// generic class name grammar:
 			//
 			// mangled(open_generic_type_name) "_$$$_" M(parameter_class_name) ( "_$$_" M(parameter_class_name) )* "_$$$$_"
@@ -971,6 +967,14 @@ namespace IKVM.Internal
 			{
 				return wrapper;
 			}
+#if STUB_GENERATOR
+			if(type.__IsMissing || type.__ContainsMissingType)
+			{
+				wrapper = new UnloadableTypeWrapper("Missing/" + type.Assembly.FullName);
+				globalTypeToTypeWrapper.Add(type, wrapper);
+				return wrapper;
+			}
+#endif
 			string remapped;
 			if(remappedTypes.TryGetValue(type, out remapped))
 			{
@@ -1204,6 +1208,14 @@ namespace IKVM.Internal
 
 		internal static TypeWrapper LoadClassCritical(string name)
 		{
+#if STATIC_COMPILER
+			TypeWrapper wrapper = GetBootstrapClassLoader().LoadClassByDottedNameFast(name);
+			if (wrapper == null)
+			{
+				throw new FatalCompilerErrorException(Message.CriticalClassNotFound, name);
+			}
+			return wrapper;
+#else
 			try
 			{
 				return GetBootstrapClassLoader().LoadClassByDottedName(name);
@@ -1213,6 +1225,7 @@ namespace IKVM.Internal
 				JVM.CriticalFailure("Loading of critical class failed", x);
 				return null;
 			}
+#endif
 		}
 
 		internal void RegisterNativeLibrary(IntPtr p)
@@ -1349,14 +1362,14 @@ namespace IKVM.Internal
 #if STATIC_COMPILER
 		internal virtual void IssueMessage(Message msgId, params string[] values)
 		{
-			// only code that is compiled (by the static compiler) can generate warnings
-			// (so that must mean an instance of CompilerClassLoader, which overrides this method.)
-			throw new InvalidOperationException();
+			// it's not ideal when we end up here (because it means we're emitting a warning that is not associated with a specific output target),
+			// but it happens when we're decoding something in a referenced assembly that either doesn't make sense or contains an unloadable type
+			StaticCompiler.IssueMessage(msgId, values);
 		}
 #endif
 	}
 
-	class GenericClassLoader : ClassLoaderWrapper
+	sealed class GenericClassLoader : ClassLoaderWrapper
 	{
 		private ClassLoaderWrapper[] delegates;
 
