@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2011 Jeroen Frijters
+  Copyright (C) 2002-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -53,9 +53,11 @@ namespace IKVM.Internal
 		internal const string GenericAttributeAnnotationMultipleTypeName = "ikvm.internal.AttributeAnnotationMultiple`1";
 		private static readonly Dictionary<Type, TypeWrapper> types = new Dictionary<Type, TypeWrapper>();
 		private readonly Type type;
+		private TypeWrapper baseTypeWrapper;
 		private TypeWrapper[] innerClasses;
 		private TypeWrapper outerClass;
 		private TypeWrapper[] interfaces;
+		private volatile bool finished;
 
 		private static Modifiers GetModifiers(Type type)
 		{
@@ -301,11 +303,16 @@ namespace IKVM.Internal
 			}
 
 			internal OpenGenericTypeWrapper(Type type, string name)
-				: base(GetModifiers(type), name, type.IsInterface ? null : CoreClasses.java.lang.Object.Wrapper)
+				: base(GetModifiers(type), name)
 			{
 				this.type = type;
 			}
 
+			internal override TypeWrapper BaseTypeWrapper
+			{
+				get { return type.IsInterface ? null : CoreClasses.java.lang.Object.Wrapper; }
+			}
+
 			internal override TypeWrapper DeclaringTypeWrapper
 			{
 				get { return null; }
@@ -344,9 +351,17 @@ namespace IKVM.Internal
 
 		internal abstract class FakeTypeWrapper : TypeWrapper
 		{
+			private readonly TypeWrapper baseWrapper;
+
 			protected FakeTypeWrapper(Modifiers modifiers, string name, TypeWrapper baseWrapper)
-				: base(modifiers, name, baseWrapper)
+				: base(modifiers, name)
 			{
+				this.baseWrapper = baseWrapper;
+			}
+
+			internal override TypeWrapper BaseTypeWrapper
+			{
+				get { return baseWrapper; }
 			}
 
 			internal sealed override bool IsFakeNestedType
@@ -371,14 +386,21 @@ namespace IKVM.Internal
 				ParameterInfo[] parameters = invoke.GetParameters();
 				TypeWrapper[] argTypeWrappers = new TypeWrapper[parameters.Length];
 				System.Text.StringBuilder sb = new System.Text.StringBuilder("(");
+				MemberFlags flags = MemberFlags.None;
 				for (int i = 0; i < parameters.Length; i++)
 				{
-					argTypeWrappers[i] = ClassLoaderWrapper.GetWrapperFromType(parameters[i].ParameterType);
+					Type parameterType = parameters[i].ParameterType;
+					if (parameterType.IsByRef)
+					{
+						flags |= MemberFlags.DelegateInvokeWithByRefParameter;
+						parameterType = ArrayTypeWrapper.MakeArrayType(parameterType.GetElementType(), 1);
+					}
+					argTypeWrappers[i] = ClassLoaderWrapper.GetWrapperFromType(parameterType);
 					sb.Append(argTypeWrappers[i].SigName);
 				}
 				TypeWrapper returnType = ClassLoaderWrapper.GetWrapperFromType(invoke.ReturnType);
 				sb.Append(")").Append(returnType.SigName);
-				MethodWrapper invokeMethod = new DynamicOnlyMethodWrapper(this, "Invoke", sb.ToString(), returnType, argTypeWrappers);
+				MethodWrapper invokeMethod = new DynamicOnlyMethodWrapper(this, "Invoke", sb.ToString(), returnType, argTypeWrappers, flags);
 				SetMethods(new MethodWrapper[] { invokeMethod });
 				SetFields(FieldWrapper.EmptyArray);
 			}
@@ -432,8 +454,8 @@ namespace IKVM.Internal
 
 		private class DynamicOnlyMethodWrapper : MethodWrapper, ICustomInvoke
 		{
-			internal DynamicOnlyMethodWrapper(TypeWrapper declaringType, string name, string sig, TypeWrapper returnType, TypeWrapper[] parameterTypes)
-				: base(declaringType, name, sig, null, returnType, parameterTypes, Modifiers.Public | Modifiers.Abstract, MemberFlags.None)
+			internal DynamicOnlyMethodWrapper(TypeWrapper declaringType, string name, string sig, TypeWrapper returnType, TypeWrapper[] parameterTypes, MemberFlags flags)
+				: base(declaringType, name, sig, null, returnType, parameterTypes, Modifiers.Public | Modifiers.Abstract, flags)
 			{
 			}
 
@@ -452,10 +474,14 @@ namespace IKVM.Internal
 				// we can resolve to a real method and call that instead
 				TypeWrapper tw = TypeWrapper.FromClass(NativeCode.ikvm.runtime.Util.getClassFromObject(obj));
 				MethodWrapper mw = tw.GetMethodWrapper(this.Name, this.Signature, true);
-				if (mw == null)
+				if (mw == null || mw.IsStatic)
 				{
 					throw new java.lang.AbstractMethodError(tw.Name + "." + this.Name + this.Signature);
 				}
+				if (!mw.IsPublic)
+				{
+					throw new java.lang.IllegalAccessError(tw.Name + "." + this.Name + this.Signature);
+				}
 				java.lang.reflect.Method m = (java.lang.reflect.Method)mw.ToMethodOrConstructor(true);
 				m.@override = true;
 				return m.invoke(obj, args, callerID);
@@ -484,7 +510,7 @@ namespace IKVM.Internal
 			}
 #endif
 
-			private class EnumFieldWrapper : FieldWrapper
+			private sealed class EnumFieldWrapper : FieldWrapper
 			{
 #if !STATIC_COMPILER && !STUB_GENERATOR
 				private readonly int ordinal;
@@ -528,7 +554,7 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 			}
 
-			private class EnumValuesMethodWrapper : MethodWrapper, ICustomInvoke
+			private sealed class EnumValuesMethodWrapper : MethodWrapper, ICustomInvoke
 			{
 				internal EnumValuesMethodWrapper(TypeWrapper declaringType)
 					: base(declaringType, "values", "()[" + declaringType.SigName, null, declaringType.MakeArrayType(1), TypeWrapper.EmptyArray, Modifiers.Public | Modifiers.Static, MemberFlags.None)
@@ -557,7 +583,7 @@ namespace IKVM.Internal
 #endif // !STATIC_COMPILER && !FIRST_PASS && !STUB_GENERATOR
 			}
 
-			private class EnumValueOfMethodWrapper : MethodWrapper, ICustomInvoke
+			private sealed class EnumValueOfMethodWrapper : MethodWrapper, ICustomInvoke
 			{
 				internal EnumValueOfMethodWrapper(TypeWrapper declaringType)
 					: base(declaringType, "valueOf", "(Ljava.lang.String;)" + declaringType.SigName, null, declaringType, new TypeWrapper[] { CoreClasses.java.lang.String.Wrapper }, Modifiers.Public | Modifiers.Static, MemberFlags.None)
@@ -800,7 +826,7 @@ namespace IKVM.Internal
 				singleOneArgCtor = oneArgCtorCount == 1 ? oneArgCtor : null;
 			}
 
-			private class AttributeAnnotationMethodWrapper : DynamicOnlyMethodWrapper
+			private sealed class AttributeAnnotationMethodWrapper : DynamicOnlyMethodWrapper
 			{
 				private bool optional;
 
@@ -877,7 +903,7 @@ namespace IKVM.Internal
 				}
 
 				private AttributeAnnotationMethodWrapper(AttributeAnnotationTypeWrapper tw, string name, TypeWrapper returnType, bool optional)
-					: base(tw, name, "()" + returnType.SigName, returnType, TypeWrapper.EmptyArray)
+					: base(tw, name, "()" + returnType.SigName, returnType, TypeWrapper.EmptyArray, MemberFlags.None)
 				{
 					this.optional = optional;
 				}
@@ -900,7 +926,8 @@ namespace IKVM.Internal
 				}
 				foreach (PropertyInfo pi in attributeType.GetProperties(BindingFlags.Instance | BindingFlags.Public))
 				{
-					if (pi.CanRead && pi.CanWrite && IsSupportedType(pi.PropertyType))
+					// the getter and setter methods both need to be public
+					if (pi.GetGetMethod() != null && pi.GetSetMethod() != null && IsSupportedType(pi.PropertyType))
 					{
 						methods.Add(new AttributeAnnotationMethodWrapper(this, pi.Name, pi.PropertyType, true));
 					}
@@ -1015,7 +1042,7 @@ namespace IKVM.Internal
 					{
 						tw = tw.MakeArrayType(1);
 					}
-					SetMethods(new MethodWrapper[] { new DynamicOnlyMethodWrapper(this, "value", "()" + tw.SigName, tw, TypeWrapper.EmptyArray) });
+					SetMethods(new MethodWrapper[] { new DynamicOnlyMethodWrapper(this, "value", "()" + tw.SigName, tw, TypeWrapper.EmptyArray, MemberFlags.None) });
 					SetFields(FieldWrapper.EmptyArray);
 				}
 
@@ -1057,7 +1084,7 @@ namespace IKVM.Internal
 				}
 #endif
 
-				private class ReturnValueAnnotation : Annotation
+				private sealed class ReturnValueAnnotation : Annotation
 				{
 					private AttributeAnnotationTypeWrapper type;
 
@@ -1123,6 +1150,11 @@ namespace IKVM.Internal
 					internal override void Apply(ClassLoaderWrapper loader, PropertyBuilder pb, object annotation)
 					{
 					}
+
+					internal override bool IsCustomAttribute
+					{
+						get { return type.Annotation.IsCustomAttribute; }
+					}
 				}
 
 				internal override Annotation Annotation
@@ -1158,7 +1190,7 @@ namespace IKVM.Internal
 				protected override void LazyPublishMembers()
 				{
 					TypeWrapper tw = declaringType.MakeArrayType(1);
-					SetMethods(new MethodWrapper[] { new DynamicOnlyMethodWrapper(this, "value", "()" + tw.SigName, tw, TypeWrapper.EmptyArray) });
+					SetMethods(new MethodWrapper[] { new DynamicOnlyMethodWrapper(this, "value", "()" + tw.SigName, tw, TypeWrapper.EmptyArray, MemberFlags.None) });
 					SetFields(FieldWrapper.EmptyArray);
 				}
 
@@ -1193,7 +1225,7 @@ namespace IKVM.Internal
 				}
 #endif
 
-				private class MultipleAnnotation : Annotation
+				private sealed class MultipleAnnotation : Annotation
 				{
 					private AttributeAnnotationTypeWrapper type;
 
@@ -1281,6 +1313,11 @@ namespace IKVM.Internal
 							annot.Apply(loader, pb, ann);
 						}
 					}
+
+					internal override bool IsCustomAttribute
+					{
+						get { return type.Annotation.IsCustomAttribute; }
+					}
 				}
 
 				internal override Annotation Annotation
@@ -1399,7 +1436,7 @@ namespace IKVM.Internal
 			}
 #endif
 
-			private class AttributeAnnotation : Annotation
+			private sealed class AttributeAnnotation : Annotation
 			{
 				private Type type;
 
@@ -1632,6 +1669,11 @@ namespace IKVM.Internal
 						pb.SetCustomAttribute(MakeCustomAttributeBuilder(loader, annotation));
 					}
 				}
+
+				internal override bool IsCustomAttribute
+				{
+					get { return true; }
+				}
 			}
 
 			internal override Annotation Annotation
@@ -1706,7 +1748,7 @@ namespace IKVM.Internal
 		}
 
 		private DotNetTypeWrapper(Type type, string name)
-			: base(GetModifiers(type), name, GetBaseTypeWrapper(type))
+			: base(GetModifiers(type), name)
 		{
 			Debug.Assert(!(type.IsByRef), type.FullName);
 			Debug.Assert(!(type.IsPointer), type.FullName);
@@ -1717,6 +1759,11 @@ namespace IKVM.Internal
 			this.type = type;
 		}
 
+		internal override TypeWrapper BaseTypeWrapper
+		{
+			get { return baseTypeWrapper ?? (baseTypeWrapper = GetBaseTypeWrapper(type)); }
+		}
+
 		internal override ClassLoaderWrapper GetClassLoader()
 		{
 			if (type.IsGenericType)
@@ -1734,7 +1781,22 @@ namespace IKVM.Internal
 			}
 		}
 
-		private class DelegateMethodWrapper : MethodWrapper
+		internal static string GetDelegateInvokeStubName(Type delegateType)
+		{
+			MethodInfo delegateInvoke = delegateType.GetMethod("Invoke");
+			ParameterInfo[] parameters = delegateInvoke.GetParameters();
+			string name = null;
+			for (int i = 0; i < parameters.Length; i++)
+			{
+				if (parameters[i].ParameterType.IsByRef)
+				{
+					name = (name ?? "<Invoke>") + "_" + i;
+				}
+			}
+			return name ?? "Invoke";
+		}
+
+		private sealed class DelegateMethodWrapper : MethodWrapper
 		{
 			private ConstructorInfo delegateConstructor;
 			private DelegateInnerClassTypeWrapper iface;
@@ -1754,33 +1816,53 @@ namespace IKVM.Internal
 				{
 					return false;
 				}
-				context.Emitter.Emit(OpCodes.Dup);
 				// we know that a DelegateInnerClassTypeWrapper has only one method
 				Debug.Assert(iface.GetMethods().Length == 1);
-				MethodWrapper mw = targetType.GetMethodWrapper("Invoke", iface.GetMethods()[0].Signature, true);
+				MethodWrapper mw = targetType.GetMethodWrapper(GetDelegateInvokeStubName(DeclaringType.TypeAsTBD), iface.GetMethods()[0].Signature, true);
+				if (mw == null || mw.IsStatic || !mw.IsPublic)
+				{
+					context.Emitter.Emit(OpCodes.Ldftn, CreateErrorStub(context, targetType, mw == null || mw.IsStatic));
+					context.Emitter.Emit(OpCodes.Newobj, delegateConstructor);
+					return true;
+				}
 				// TODO linking here is not safe
 				mw.Link();
-				context.Emitter.Emit(OpCodes.Ldvirtftn, (MethodInfo)mw.GetMethod());
+				context.Emitter.Emit(OpCodes.Dup);
+				context.Emitter.Emit(OpCodes.Ldvirtftn, mw.GetMethod());
 				context.Emitter.Emit(OpCodes.Newobj, delegateConstructor);
 				return true;
 			}
 
+			private MethodInfo CreateErrorStub(EmitIntrinsicContext context, TypeWrapper targetType, bool isAbstract)
+			{
+				MethodInfo invoke = delegateConstructor.DeclaringType.GetMethod("Invoke");
+				ParameterInfo[] parameters = invoke.GetParameters();
+				Type[] parameterTypes = new Type[parameters.Length + 1];
+				parameterTypes[0] = Types.Object;
+				for (int i = 0; i < parameters.Length; i++)
+				{
+					parameterTypes[i + 1] = parameters[i].ParameterType;
+				}
+				MethodBuilder mb = context.Context.DefineDelegateInvokeErrorStub(invoke.ReturnType, parameterTypes);
+				CodeEmitter ilgen = CodeEmitter.Create(mb);
+				ilgen.EmitThrow(isAbstract ? "java.lang.AbstractMethodError" : "java.lang.IllegalAccessError", targetType.Name + ".Invoke" + iface.GetMethods()[0].Signature);
+				ilgen.DoEmit();
+				return mb;
+			}
+
 			internal override void EmitNewobj(CodeEmitter ilgen)
 			{
-				MethodInfo createDelegate = Types.Delegate.GetMethod("CreateDelegate", new Type[] { Types.Type, Types.Object, Types.String });
-				CodeEmitterLocal targetObj = ilgen.DeclareLocal(Types.Object);
-				ilgen.Emit(OpCodes.Stloc, targetObj);
 				ilgen.Emit(OpCodes.Ldtoken, delegateConstructor.DeclaringType);
 				ilgen.Emit(OpCodes.Call, Types.Type.GetMethod("GetTypeFromHandle", new Type[] { Types.RuntimeTypeHandle }));
-				ilgen.Emit(OpCodes.Ldloc, targetObj);
-				ilgen.Emit(OpCodes.Ldstr, "Invoke");
-				ilgen.Emit(OpCodes.Call, createDelegate);
+				ilgen.Emit(OpCodes.Ldstr, GetDelegateInvokeStubName(DeclaringType.TypeAsTBD));
+				ilgen.Emit(OpCodes.Ldstr, iface.GetMethods()[0].Signature);
+				ilgen.Emit(OpCodes.Call, ByteCodeHelperMethods.DynamicCreateDelegate);
 				ilgen.Emit(OpCodes.Castclass, delegateConstructor.DeclaringType);
 			}
 #endif // !STUB_GENERATOR
 		}
 
-		private class ByRefMethodWrapper : SmartMethodWrapper
+		private sealed class ByRefMethodWrapper : SmartMethodWrapper
 		{
 #if !STATIC_COMPILER
 			private bool[] byrefs;
@@ -1799,29 +1881,23 @@ namespace IKVM.Internal
 #if !STUB_GENERATOR
 			protected override void CallImpl(CodeEmitter ilgen)
 			{
-				MethodBase mb = GetMethod();
-				MethodInfo mi = mb as MethodInfo;
-				if (mi != null)
-				{
-					ilgen.Emit(OpCodes.Call, mi);
-				}
-				else
-				{
-					ilgen.Emit(OpCodes.Call, (ConstructorInfo)mb);
-				}
+				ConvertByRefArgs(ilgen);
+				ilgen.Emit(OpCodes.Call, GetMethod());
 			}
 
 			protected override void CallvirtImpl(CodeEmitter ilgen)
 			{
-				ilgen.Emit(OpCodes.Callvirt, (MethodInfo)GetMethod());
+				ConvertByRefArgs(ilgen);
+				ilgen.Emit(OpCodes.Callvirt, GetMethod());
 			}
 
 			protected override void NewobjImpl(CodeEmitter ilgen)
 			{
-				ilgen.Emit(OpCodes.Newobj, (ConstructorInfo)GetMethod());
+				ConvertByRefArgs(ilgen);
+				ilgen.Emit(OpCodes.Newobj, GetMethod());
 			}
 
-			protected override void PreEmit(CodeEmitter ilgen)
+			private void ConvertByRefArgs(CodeEmitter ilgen)
 			{
 				CodeEmitterLocal[] locals = new CodeEmitterLocal[args.Length];
 				for (int i = args.Length - 1; i >= 0; i--)
@@ -1843,12 +1919,11 @@ namespace IKVM.Internal
 						ilgen.Emit(OpCodes.Ldelema, args[i].GetElementType());
 					}
 				}
-				base.PreEmit(ilgen);
 			}
 #endif // !STUB_GENERATOR
 		}
 
-		private class EnumWrapMethodWrapper : MethodWrapper
+		private sealed class EnumWrapMethodWrapper : MethodWrapper
 		{
 			internal EnumWrapMethodWrapper(DotNetTypeWrapper tw, TypeWrapper fieldType)
 				: base(tw, "wrap", "(" + fieldType.SigName + ")" + tw.SigName, null, tw, new TypeWrapper[] { fieldType }, Modifiers.Static | Modifiers.Public, MemberFlags.None)
@@ -1865,7 +1940,7 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 		}
 
-		internal class EnumValueFieldWrapper : FieldWrapper
+		internal sealed class EnumValueFieldWrapper : FieldWrapper
 		{
 			private Type underlyingType;
 
@@ -1896,7 +1971,7 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 		}
 
-		private class ValueTypeDefaultCtor : MethodWrapper
+		private sealed class ValueTypeDefaultCtor : MethodWrapper
 		{
 			internal ValueTypeDefaultCtor(DotNetTypeWrapper tw)
 				: base(tw, "<init>", "()V", null, PrimitiveTypeWrapper.VOID, TypeWrapper.EmptyArray, Modifiers.Public, MemberFlags.None)
@@ -1913,7 +1988,7 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 		}
 
-		private class FinalizeMethodWrapper : MethodWrapper
+		private sealed class FinalizeMethodWrapper : MethodWrapper
 		{
 			internal FinalizeMethodWrapper(DotNetTypeWrapper tw)
 				: base(tw, "finalize", "()V", null, PrimitiveTypeWrapper.VOID, TypeWrapper.EmptyArray, Modifiers.Protected | Modifiers.Final, MemberFlags.None)
@@ -1933,7 +2008,7 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 		}
 
-		private class CloneMethodWrapper : MethodWrapper
+		private sealed class CloneMethodWrapper : MethodWrapper
 		{
 			internal CloneMethodWrapper(DotNetTypeWrapper tw)
 				: base(tw, "clone", "()Ljava.lang.Object;", null, CoreClasses.java.lang.Object.Wrapper, TypeWrapper.EmptyArray, Modifiers.Protected | Modifiers.Final, MemberFlags.None)
@@ -1946,9 +2021,9 @@ namespace IKVM.Internal
 				ilgen.Emit(OpCodes.Dup);
 				ilgen.Emit(OpCodes.Isinst, ClassLoaderWrapper.LoadClassCritical("java.lang.Cloneable").TypeAsBaseType);
 				CodeEmitterLabel label1 = ilgen.DefineLabel();
-				ilgen.Emit(OpCodes.Brtrue_S, label1);
+				ilgen.EmitBrtrue(label1);
 				CodeEmitterLabel label2 = ilgen.DefineLabel();
-				ilgen.Emit(OpCodes.Brfalse_S, label2);
+				ilgen.EmitBrfalse(label2);
 				ilgen.EmitThrow("java.lang.CloneNotSupportedException");
 				ilgen.MarkLabel(label2);
 				ilgen.EmitThrow("java.lang.NullPointerException");
@@ -2136,8 +2211,9 @@ namespace IKVM.Internal
 							InterfaceMapping map = type.GetInterfaceMap(interfaces[i]);
 							for (int j = 0; j < map.InterfaceMethods.Length; j++)
 							{
-								if ((!map.TargetMethods[j].IsPublic || map.TargetMethods[j].Name != map.InterfaceMethods[j].Name)
-									&& map.TargetMethods[j].DeclaringType == type)
+								if (map.TargetMethods[j] == null
+									|| ((!map.TargetMethods[j].IsPublic || map.TargetMethods[j].Name != map.InterfaceMethods[j].Name)
+										&& map.TargetMethods[j].DeclaringType == type))
 								{
 									string name;
 									string sig;
@@ -2158,8 +2234,6 @@ namespace IKVM.Internal
 				// methods, which don't really exist).
 				if (ClassLoaderWrapper.IsRemappedType(type) && !type.IsSealed && !type.IsInterface)
 				{
-					// Finish the type, to make sure the methods are populated
-					this.BaseTypeWrapper.Finish();
 					TypeWrapper baseTypeWrapper = this.BaseTypeWrapper;
 					while (baseTypeWrapper != null)
 					{
@@ -2237,16 +2311,20 @@ namespace IKVM.Internal
 			}
 		}
 
-		private class BaseFinalMethodWrapper : MethodWrapper
+		private sealed class BaseFinalMethodWrapper : MethodWrapper
 		{
 			private MethodWrapper m;
 
 			internal BaseFinalMethodWrapper(DotNetTypeWrapper tw, MethodWrapper m)
-				: base(tw, m.Name, m.Signature, m.GetMethod(), m.ReturnType, m.GetParameters(), (m.Modifiers & ~Modifiers.Abstract) | Modifiers.Final, MemberFlags.None)
+				: base(tw, m.Name, m.Signature, null, null, null, (m.Modifiers & ~Modifiers.Abstract) | Modifiers.Final, MemberFlags.None)
 			{
 				this.m = m;
 			}
 
+			protected override void DoLinkMethod()
+			{
+			}
+
 #if !STUB_GENERATOR
 			internal override void EmitCall(CodeEmitter ilgen)
 			{
@@ -2267,7 +2345,7 @@ namespace IKVM.Internal
 			if (mb.IsAbstract)
 			{
 				MethodInfo mi = (MethodInfo)mb;
-				if (mi.ReturnType.IsByRef || IsPointerType(mi.ReturnType))
+				if (mi.ReturnType.IsByRef || IsPointerType(mi.ReturnType) || mb.IsGenericMethodDefinition)
 				{
 					return true;
 				}
@@ -2422,8 +2500,8 @@ namespace IKVM.Internal
 				{
 					foreach (ParameterInfo p in invoke.GetParameters())
 					{
-						// TODO at the moment we don't support delegates with pointer or byref parameters
-						if (p.ParameterType.IsPointer || p.ParameterType.IsByRef)
+						// we don't support delegates with pointer parameters
+						if (p.ParameterType.IsPointer)
 						{
 							return false;
 						}
@@ -2577,14 +2655,7 @@ namespace IKVM.Internal
 			}
 			else
 			{
-				if (mb is ConstructorInfo)
-				{
-					return new SmartConstructorMethodWrapper(this, name, sig, (ConstructorInfo)mb, argTypeWrappers, mods, MemberFlags.None);
-				}
-				else
-				{
-					return new SmartCallMethodWrapper(this, name, sig, (MethodInfo)mb, retTypeWrapper, argTypeWrappers, mods, MemberFlags.None, SimpleOpCode.Call, SimpleOpCode.Callvirt);
-				}
+				return new TypicalMethodWrapper(this, name, sig, mb, retTypeWrapper, argTypeWrappers, mods, MemberFlags.None);
 			}
 		}
 
@@ -2638,6 +2709,11 @@ namespace IKVM.Internal
 
 		internal override void Finish()
 		{
+			// we don't need locking, because Finish and Link are idempotent
+			if (finished)
+			{
+				return;
+			}
 			if (BaseTypeWrapper != null)
 			{
 				BaseTypeWrapper.Finish();
@@ -2646,6 +2722,15 @@ namespace IKVM.Internal
 			{
 				tw.Finish();
 			}
+			foreach (MethodWrapper mw in GetMethods())
+			{
+				mw.Link();
+			}
+			foreach (FieldWrapper fw in GetFields())
+			{
+				fw.Link();
+			}
+			finished = true;
 		}
 
 #if !STATIC_COMPILER && !STUB_GENERATOR
@@ -2693,7 +2778,7 @@ namespace IKVM.Internal
 
 		internal override bool IsFastClassLiteralSafe
 		{
-			get { return true; }
+			get { return type != Types.Void && !type.IsPrimitive && !IsRemapped; }
 		}
 	}
 }
