@@ -37,9 +37,6 @@ namespace IKVM.Internal
 {
 	sealed class DynamicClassLoader : TypeWrapperFactory
 	{
-		// note that MangleNestedTypeName() assumes that there are less than 16 special characters
-		private static readonly char[] specialCharacters = { '\\', '+', ',', '[', ']', '*', '&', '\u0000' };
-		private static readonly string specialCharactersString = new String(specialCharacters);
 #if !STATIC_COMPILER
 		private static List<AssemblyBuilder> saveDebugAssemblies;
 		private static List<DynamicClassLoader> saveClassLoaders;
@@ -57,6 +54,9 @@ namespace IKVM.Internal
 #if !STATIC_COMPILER && !CLASSGC
 		private static DynamicClassLoader instance = new DynamicClassLoader(CreateModuleBuilder());
 #endif
+#if CLASSGC
+		private List<string> friends = new List<string>();
+#endif
 
 		[System.Security.SecuritySafeCritical]
 		static DynamicClassLoader()
@@ -83,6 +83,21 @@ namespace IKVM.Internal
 			dynamicTypes.Add("<Module>", null);
 		}
 
+#if CLASSGC
+		internal override void AddInternalsVisibleTo(Assembly friend)
+		{
+			string name = friend.GetName().Name;
+			lock (friends)
+			{
+				if (!friends.Contains(name))
+				{
+					friends.Add(name);
+					((AssemblyBuilder)moduleBuilder.Assembly).SetCustomAttribute(new CustomAttributeBuilder(typeof(System.Runtime.CompilerServices.InternalsVisibleToAttribute).GetConstructor(new Type[] { typeof(string) }), new object[] { name }));
+				}
+			}
+		}
+#endif // CLASSGC
+
 #if !STATIC_COMPILER
 		private static Assembly OnTypeResolve(object sender, ResolveEventArgs args)
 		{
@@ -119,36 +134,6 @@ namespace IKVM.Internal
 		}
 #endif // !STATIC_COMPILER
 
-		internal static string EscapeName(string name)
-		{
-			// TODO the escaping of special characters is not required on .NET 2.0
-			// (but it doesn't really hurt that much either, the only overhead is the
-			// extra InnerClassAttribute to record the real name of the class)
-			// Note that even though .NET 2.0 automatically escapes the special characters,
-			// the name that gets passed in ResolveEventArgs.Name of the TypeResolve event
-			// contains the unescaped type name.
-			if(name.IndexOfAny(specialCharacters) >= 0)
-			{
-				System.Text.StringBuilder sb = new System.Text.StringBuilder();
-				foreach(char c in name)
-				{
-					if(specialCharactersString.IndexOf(c) >= 0)
-					{
-						if(c == 0)
-						{
-							// we can't escape the NUL character, so we replace it with a space.
-							sb.Append(' ');
-							continue;
-						}
-						sb.Append('\\');
-					}
-					sb.Append(c);
-				}
-				name = sb.ToString();
-			}
-			return name;
-		}
-
 		internal override bool ReserveName(string name)
 		{
 			lock(dynamicTypes)
@@ -162,49 +147,60 @@ namespace IKVM.Internal
 			}
 		}
 
-		private string AllocMangledName(string mangledTypeName)
+		internal override string AllocMangledName(DynamicTypeWrapper tw)
 		{
 			lock(dynamicTypes)
 			{
-				mangledTypeName = EscapeName(mangledTypeName);
-				// FXBUG the CLR (both 1.1 and 2.0) doesn't like type names that end with a single period,
-				// it loses the trailing period in the name that gets passed in the TypeResolve event.
-				if(dynamicTypes.ContainsKey(mangledTypeName) || mangledTypeName.EndsWith("."))
-				{
+				return TypeNameMangleImpl(dynamicTypes, tw.Name, tw);
+			}
+		}
+
+		internal static string TypeNameMangleImpl(Dictionary<string, TypeWrapper> dict, string name, TypeWrapper tw)
+		{
+			// the CLR maximum type name length is 1023 characters,
+			// but we need to leave some room for the suffix that we
+			// may need to append to make the name unique
+			const int MaxLength = 1000;
+			if (name.Length > MaxLength)
+			{
+				name = name.Substring(0, MaxLength) + "/truncated";
+			}
+			string mangledTypeName = TypeNameUtil.ReplaceIllegalCharacters(name);
+			// FXBUG the CLR (both 1.1 and 2.0) doesn't like type names that end with a single period,
+			// it loses the trailing period in the name that gets passed in the TypeResolve event.
+			if (dict.ContainsKey(mangledTypeName) || mangledTypeName.EndsWith("."))
+			{
 #if STATIC_COMPILER
 					Tracer.Warning(Tracer.Compiler, "Class name clash: {0}", mangledTypeName);
 #endif
-					// Java class names cannot contain slashes (since they are converted into periods),
-					// so we take advantage of that fact to create a unique name.
-					string baseName = mangledTypeName;
-					int instanceId = 0;
-					do
-					{
-						mangledTypeName = baseName + "/" + (++instanceId);
-					} while(dynamicTypes.ContainsKey(mangledTypeName));
-				}
-				dynamicTypes.Add(mangledTypeName, null);
+				// Java class names cannot contain slashes (since they are converted into periods),
+				// so we take advantage of that fact to create a unique name.
+				string baseName = mangledTypeName;
+				int instanceId = 0;
+				do
+				{
+					mangledTypeName = baseName + "/" + (++instanceId);
+				} while (dict.ContainsKey(mangledTypeName));
 			}
+			dict.Add(mangledTypeName, tw);
 			return mangledTypeName;
 		}
 
 		internal sealed override TypeWrapper DefineClassImpl(Dictionary<string, TypeWrapper> types, ClassFile f, ClassLoaderWrapper classLoader, object protectionDomain)
 		{
-			DynamicTypeWrapper type;
 #if STATIC_COMPILER
-			type = new AotTypeWrapper(f, (CompilerClassLoader)classLoader);
+			AotTypeWrapper type = new AotTypeWrapper(f, (CompilerClassLoader)classLoader);
+			type.CreateStep1();
+			types[f.Name] = type;
+			return type;
 #else
-			type = new DynamicTypeWrapper(f, classLoader);
-#endif
 			// this step can throw a retargettable exception, if the class is incorrect
-			bool hasclinit;
-			type.CreateStep1(out hasclinit);
-			// now we can allocate the mangledTypeName, because the next step cannot fail
-			string mangledTypeName = AllocMangledName(f.Name);
+			DynamicTypeWrapper type = new DynamicTypeWrapper(f, classLoader);
 			// This step actually creates the TypeBuilder. It is not allowed to throw any exceptions,
 			// if an exception does occur, it is due to a programming error in the IKVM or CLR runtime
 			// and will cause a CriticalFailure and exit the process.
-			type.CreateStep2NoFail(hasclinit, mangledTypeName);
+			type.CreateStep1();
+			type.CreateStep2();
 			lock(types)
 			{
 				// in very extreme conditions another thread may have beaten us to it
@@ -215,13 +211,8 @@ namespace IKVM.Internal
 				types.TryGetValue(f.Name, out race);
 				if(race == null)
 				{
-					lock(dynamicTypes)
-					{
-						Debug.Assert(dynamicTypes.ContainsKey(mangledTypeName) && dynamicTypes[mangledTypeName] == null);
-						dynamicTypes[mangledTypeName] = type;
-					}
 					types[f.Name] = type;
-#if !STATIC_COMPILER && !FIRST_PASS
+#if !FIRST_PASS
 					java.lang.Class clazz = new java.lang.Class(null);
 #if __MonoCS__
 					TypeWrapper.SetTypeWrapperHack(clazz, type);
@@ -238,6 +229,7 @@ namespace IKVM.Internal
 				}
 			}
 			return type;
+#endif // STATIC_COMPILER
 		}
 
 #if STATIC_COMPILER
@@ -250,7 +242,7 @@ namespace IKVM.Internal
 				AttributeHelper.SetEditorBrowsableNever(proxyHelperContainer);
 				proxyHelpers = new List<TypeBuilder>();
 			}
-			proxyHelpers.Add(proxyHelperContainer.DefineNestedType(MangleNestedTypeName(type.FullName), TypeAttributes.NestedPublic | TypeAttributes.Interface | TypeAttributes.Abstract, null, new Type[] { type }));
+			proxyHelpers.Add(proxyHelperContainer.DefineNestedType(TypeNameUtil.MangleNestedTypeName(type.FullName), TypeAttributes.NestedPublic | TypeAttributes.Interface | TypeAttributes.Abstract, null, new Type[] { type }));
 		}
 
 		internal TypeBuilder DefineProxy(TypeWrapper proxyClass, TypeWrapper[] interfaces)
@@ -280,7 +272,7 @@ namespace IKVM.Internal
 			{
 				sb.Append(tw.Name.Length).Append('|').Append(tw.Name);
 			}
-			return MangleNestedTypeName(sb.ToString());
+			return TypeNameUtil.MangleNestedTypeName(sb.ToString());
 		}
 
 		internal static string GetProxyName(TypeWrapper[] interfaces)
@@ -290,37 +282,7 @@ namespace IKVM.Internal
 
 		internal static string GetProxyHelperName(Type type)
 		{
-			return "__<Proxy>+" + MangleNestedTypeName(type.FullName);
-		}
-
-		private static string MangleNestedTypeName(string name)
-		{
-			System.Text.StringBuilder sb = new System.Text.StringBuilder();
-			foreach (char c in name)
-			{
-				int index = specialCharactersString.IndexOf(c);
-				if(c == '.')
-				{
-					sb.Append("_");
-				}
-				else if(c == '_')
-				{
-					sb.Append("^-");
-				}
-				else if(index == -1)
-				{
-					sb.Append(c);
-					if(c == '^')
-					{
-						sb.Append(c);
-					}
-				}
-				else
-				{
-					sb.Append('^').AppendFormat("{0:X1}", index);
-				}
-			}
-			return sb.ToString();
+			return "__<Proxy>+" + TypeNameUtil.MangleNestedTypeName(type.FullName);
 		}
 
 		internal override Type DefineUnloadable(string name)
@@ -341,7 +303,7 @@ namespace IKVM.Internal
 					unloadableContainer = moduleBuilder.DefineType("__<Unloadable>", TypeAttributes.Interface | TypeAttributes.Abstract);
 					AttributeHelper.HideFromJava(unloadableContainer);
 				}
-				type = unloadableContainer.DefineNestedType(MangleNestedTypeName(name), TypeAttributes.NestedPrivate | TypeAttributes.Interface | TypeAttributes.Abstract);
+				type = unloadableContainer.DefineNestedType(TypeNameUtil.MangleNestedTypeName(name), TypeAttributes.NestedPrivate | TypeAttributes.Interface | TypeAttributes.Abstract);
 				unloadables.Add(name, type);
 				return type;
 			}
