@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002-2009 Jeroen Frijters
+  Copyright (C) 2002-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -50,10 +50,12 @@ namespace IKVM.Internal
 		PropertyAccessor = 64,
 		Intrinsic = 128,
 		CallerID = 256,
-		NonPublicTypeInSignature = 512,	// this flag is only available after linking
+		NonPublicTypeInSignature = 512,	// this flag is only available after linking and is not set for access stubs
+		DelegateInvokeWithByRefParameter = 1024,
+		Type2FinalField = 2048,
 	}
 
-	class MemberWrapper
+	abstract class MemberWrapper
 	{
 		private HandleWrapper handle;
 		private TypeWrapper declaringType;
@@ -160,7 +162,7 @@ namespace IKVM.Internal
 					// The JVM supports accessing members that have non-public types in their signature from another package,
 					// but the CLI doesn't. It would be nice if we worked around that by emitting extra accessors, but for now
 					// we'll simply disallow such access across assemblies (unless the appropriate InternalsVisibleToAttribute exists).
-					&& ((flags & MemberFlags.NonPublicTypeInSignature) == 0 || InPracticeInternalsVisibleTo(caller));
+					&& (!(HasNonPublicTypeInSignature || IsType2FinalField) || InPracticeInternalsVisibleTo(caller));
 			}
 			return false;
 		}
@@ -184,6 +186,14 @@ namespace IKVM.Internal
 				// (this handles the case where we're running in dynamic mode)
 				return true;
 			}
+#endif
+#if CLASSGC
+			if (DeclaringType is DynamicTypeWrapper)
+			{
+				// if we are dynamic, we can just become friends with the caller
+				DeclaringType.GetClassLoader().GetTypeWrapperFactory().AddInternalsVisibleTo(caller.TypeAsTBD.Assembly);
+				return true;
+			}
 #endif
 			return DeclaringType.InternalsVisibleTo(caller);
 		}
@@ -258,6 +268,21 @@ namespace IKVM.Internal
 			flags |= MemberFlags.NonPublicTypeInSignature;
 		}
 
+		internal bool HasNonPublicTypeInSignature
+		{
+			get { return (flags & MemberFlags.NonPublicTypeInSignature) != 0; }
+		}
+
+		protected void SetType2FinalField()
+		{
+			flags |= MemberFlags.Type2FinalField;
+		}
+	
+		internal bool IsType2FinalField
+		{
+			get { return (flags & MemberFlags.Type2FinalField) != 0; }
+		}
+
 		internal bool HasCallerID
 		{
 			get
@@ -266,6 +291,11 @@ namespace IKVM.Internal
 			}
 		}
 
+		internal bool IsDelegateInvokeWithByRefParameter
+		{
+			get { return (flags & MemberFlags.DelegateInvokeWithByRefParameter) != 0; }
+		}
+
 		internal Modifiers Modifiers
 		{
 			get
@@ -377,66 +407,6 @@ namespace IKVM.Internal
 			}
 		}
 
-		internal class GhostMethodWrapper : SmartMethodWrapper
-		{
-			private MethodInfo ghostMethod;
-
-			internal GhostMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodBase method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
-				: base(declaringType, name, sig, method, returnType, parameterTypes, modifiers, flags)
-			{
-				// make sure we weren't handed the ghostMethod in the wrapper value type
-				Debug.Assert(method == null || method.DeclaringType.IsInterface);
-			}
-
-			private void ResolveGhostMethod()
-			{
-				if(ghostMethod == null)
-				{
-					ghostMethod = DeclaringType.TypeAsSignatureType.GetMethod(this.Name, this.GetParametersForDefineMethod());
-					if(ghostMethod == null)
-					{
-						throw new InvalidOperationException("Unable to resolve ghost method");
-					}
-				}
-			}
-
-#if !STUB_GENERATOR
-			protected override void CallvirtImpl(CodeEmitter ilgen)
-			{
-				ResolveGhostMethod();
-				ilgen.Emit(OpCodes.Call, ghostMethod);
-			}
-#endif
-		}
-
-		internal static MethodWrapper Create(TypeWrapper declaringType, string name, string sig, MethodBase method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
-		{
-			Debug.Assert(declaringType != null && name!= null && sig != null && method != null);
-
-			if(declaringType.IsGhost)
-			{
-				// HACK since our caller isn't aware of the ghost issues, we'll handle the method swapping
-				if(method.DeclaringType.IsValueType)
-				{
-					Type[] types = new Type[parameterTypes.Length];
-					for(int i = 0; i < types.Length; i++)
-					{
-						types[i] = parameterTypes[i].TypeAsSignatureType;
-					}
-					method = declaringType.TypeAsBaseType.GetMethod(method.Name, types);
-				}
-				return new GhostMethodWrapper(declaringType, name, sig, method, returnType, parameterTypes, modifiers, flags);
-			}
-			else if(method is ConstructorInfo)
-			{
-				return new SmartConstructorMethodWrapper(declaringType, name, sig, (ConstructorInfo)method, parameterTypes, modifiers, flags);
-			}
-			else
-			{
-				return new SmartCallMethodWrapper(declaringType, name, sig, (MethodInfo)method, returnType, parameterTypes, modifiers, flags, SimpleOpCode.Call, method.IsStatic ? SimpleOpCode.Call : SimpleOpCode.Callvirt);
-			}
-		}
-
 		internal MethodWrapper(TypeWrapper declaringType, string name, string sig, MethodBase method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
 			: base(declaringType, name, sig, modifiers, flags)
 		{
@@ -454,7 +424,7 @@ namespace IKVM.Internal
 
 		private void UpdateNonPublicTypeInSignatureFlag()
 		{
-			if ((IsPublic || IsProtected) && (returnTypeWrapper != null && parameterTypeWrappers != null))
+			if ((IsPublic || IsProtected) && (returnTypeWrapper != null && parameterTypeWrappers != null) && !(this is AccessStubMethodWrapper) && !(this is AccessStubConstructorMethodWrapper))
 			{
 				if (!returnTypeWrapper.IsPublic && !returnTypeWrapper.IsUnloadable)
 				{
@@ -497,11 +467,13 @@ namespace IKVM.Internal
 			object method = reflectionMethod;
 			if (method == null)
 			{
+				Link();
+				ClassLoaderWrapper loader = this.DeclaringType.GetClassLoader();
 				TypeWrapper[] argTypes = GetParameters();
 				java.lang.Class[] parameterTypes = new java.lang.Class[argTypes.Length];
 				for (int i = 0; i < argTypes.Length; i++)
 				{
-					parameterTypes[i] = argTypes[i].ClassObject;
+					parameterTypes[i] = argTypes[i].EnsureLoadable(loader).ClassObject;
 				}
 				java.lang.Class[] checkedExceptions = GetExceptions();
 				if (this.Name == StringConstants.INIT)
@@ -523,7 +495,7 @@ namespace IKVM.Internal
 						this.DeclaringType.ClassObject,
 						this.Name,
 						parameterTypes,
-						this.ReturnType.ClassObject,
+						this.ReturnType.EnsureLoadable(loader).ClassObject,
 						checkedExceptions,
 						(int)this.Modifiers | (this.IsInternal ? 0x40000000 : 0),
 						Array.IndexOf(this.DeclaringType.GetMethods(), this),
@@ -705,6 +677,13 @@ namespace IKVM.Internal
 			return parameterTypeWrappers;
 		}
 
+#if !STUB_GENERATOR
+		internal DefineMethodHelper GetDefineMethodHelper()
+		{
+			return new DefineMethodHelper(this);
+		}
+#endif
+
 		internal Type ReturnTypeForDefineMethod
 		{
 			get
@@ -820,7 +799,7 @@ namespace IKVM.Internal
 					}
 					return acc.newInstance(args);
 				}
-				else if (method is MethodInfo)
+				else if (!ReflectUtil.IsConstructor(method))
 				{
 					Debug.Assert(method.IsStatic);
 					// we're dealing with a constructor on a remapped type, if obj is supplied, it means
@@ -1050,7 +1029,7 @@ namespace IKVM.Internal
 		}
 	}
 
-	class SmartMethodWrapper : MethodWrapper
+	abstract class SmartMethodWrapper : MethodWrapper
 	{
 		internal SmartMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodBase method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
 			: base(declaringType, name, sig, method, returnType, parameterTypes, modifiers, flags)
@@ -1058,14 +1037,9 @@ namespace IKVM.Internal
 		}
 
 #if !STUB_GENERATOR
-		protected virtual void PreEmit(CodeEmitter ilgen)
-		{
-		}
-
 		internal sealed override void EmitCall(CodeEmitter ilgen)
 		{
 			AssertLinked();
-			PreEmit(ilgen);
 			CallImpl(ilgen);
 		}
 
@@ -1077,11 +1051,10 @@ namespace IKVM.Internal
 		internal sealed override void EmitCallvirt(CodeEmitter ilgen)
 		{
 			AssertLinked();
-			PreEmit(ilgen);
 			if(DeclaringType.IsNonPrimitiveValueType)
 			{
 				// callvirt isn't allowed on a value type
-				// TODO we need to check for a null reference
+				// (we don't need to check for a null reference, because we're always dealing with an unboxed value)
 				CallImpl(ilgen);
 			}
 			else
@@ -1098,7 +1071,6 @@ namespace IKVM.Internal
 		internal sealed override void EmitNewobj(CodeEmitter ilgen)
 		{
 			AssertLinked();
-			PreEmit(ilgen);
 			NewobjImpl(ilgen);
 			if(DeclaringType.IsNonPrimitiveValueType)
 			{
@@ -1135,57 +1107,118 @@ namespace IKVM.Internal
 #if !STUB_GENERATOR
 		internal override void EmitCall(CodeEmitter ilgen)
 		{
-			ilgen.Emit(SimpleOpCodeToOpCode(call), (MethodInfo)GetMethod());
+			ilgen.Emit(SimpleOpCodeToOpCode(call), GetMethod());
 		}
 
 		internal override void EmitCallvirt(CodeEmitter ilgen)
 		{
-			ilgen.Emit(SimpleOpCodeToOpCode(callvirt), (MethodInfo)GetMethod());
+			ilgen.Emit(SimpleOpCodeToOpCode(callvirt), GetMethod());
 		}
 #endif // !STUB_GENERATOR
 	}
 
-	sealed class SmartCallMethodWrapper : SmartMethodWrapper
+	sealed class TypicalMethodWrapper : SmartMethodWrapper
 	{
-		private SimpleOpCode call;
-		private SimpleOpCode callvirt;
-
-		internal SmartCallMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodInfo method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags, SimpleOpCode call, SimpleOpCode callvirt)
+		internal TypicalMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodBase method, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
 			: base(declaringType, name, sig, method, returnType, parameterTypes, modifiers, flags)
 		{
-			this.call = call;
-			this.callvirt = callvirt;
 		}
 
 #if !STUB_GENERATOR
 		protected override void CallImpl(CodeEmitter ilgen)
 		{
-			ilgen.Emit(SimpleOpCodeToOpCode(call), (MethodInfo)GetMethod());
+			ilgen.Emit(OpCodes.Call, GetMethod());
 		}
 
 		protected override void CallvirtImpl(CodeEmitter ilgen)
 		{
-			ilgen.Emit(SimpleOpCodeToOpCode(callvirt), (MethodInfo)GetMethod());
+			ilgen.Emit(OpCodes.Callvirt, GetMethod());
+		}
+
+		protected override void NewobjImpl(CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Newobj, GetMethod());
 		}
 #endif // !STUB_GENERATOR
 	}
 
-	sealed class SmartConstructorMethodWrapper : SmartMethodWrapper
+	sealed class GhostMethodWrapper : SmartMethodWrapper
 	{
-		internal SmartConstructorMethodWrapper(TypeWrapper declaringType, string name, string sig, ConstructorInfo method, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
-			: base(declaringType, name, sig, method, PrimitiveTypeWrapper.VOID, parameterTypes, modifiers, flags)
+		private MethodInfo ghostMethod;
+
+		internal GhostMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodBase method, MethodInfo ghostMethod, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
+			: base(declaringType, name, sig, method, returnType, parameterTypes, modifiers, flags)
 		{
+			// make sure we weren't handed the ghostMethod in the wrapper value type
+			Debug.Assert(method == null || method.DeclaringType.IsInterface);
+			this.ghostMethod = ghostMethod;
+		}
+
+		private void ResolveGhostMethod()
+		{
+			if (ghostMethod == null)
+			{
+				ghostMethod = DeclaringType.TypeAsSignatureType.GetMethod(this.Name, this.GetParametersForDefineMethod());
+				if (ghostMethod == null)
+				{
+					throw new InvalidOperationException("Unable to resolve ghost method");
+				}
+			}
+		}
+
+#if !STUB_GENERATOR
+		protected override void CallvirtImpl(CodeEmitter ilgen)
+		{
+			ResolveGhostMethod();
+			ilgen.Emit(OpCodes.Call, ghostMethod);
+		}
+#endif
+	}
+
+	sealed class AccessStubMethodWrapper : SmartMethodWrapper
+	{
+		private readonly MethodInfo stubVirtual;
+		private readonly MethodInfo stubNonVirtual;
+
+		internal AccessStubMethodWrapper(TypeWrapper declaringType, string name, string sig, MethodInfo core, MethodInfo stubVirtual, MethodInfo stubNonVirtual, TypeWrapper returnType, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
+			: base(declaringType, name, sig, core, returnType, parameterTypes, modifiers, flags)
+		{
+			this.stubVirtual = stubVirtual;
+			this.stubNonVirtual = stubNonVirtual;
 		}
 
 #if !STUB_GENERATOR
 		protected override void CallImpl(CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Call, (ConstructorInfo)GetMethod());
+			ilgen.Emit(OpCodes.Call, stubNonVirtual);
+		}
+
+		protected override void CallvirtImpl(CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Callvirt, stubVirtual);
+		}
+#endif // !STUB_GENERATOR
+	}
+
+	sealed class AccessStubConstructorMethodWrapper : SmartMethodWrapper
+	{
+		private readonly ConstructorInfo stub;
+
+		internal AccessStubConstructorMethodWrapper(TypeWrapper declaringType, string sig, ConstructorInfo core, ConstructorInfo stub, TypeWrapper[] parameterTypes, Modifiers modifiers, MemberFlags flags)
+			: base(declaringType, StringConstants.INIT, sig, core, PrimitiveTypeWrapper.VOID, parameterTypes, modifiers, flags)
+		{
+			this.stub = stub;
+		}
+
+#if !STUB_GENERATOR
+		protected override void CallImpl(CodeEmitter ilgen)
+		{
+			ilgen.Emit(OpCodes.Call, stub);
 		}
 
 		protected override void NewobjImpl(CodeEmitter ilgen)
 		{
-			ilgen.Emit(OpCodes.Newobj, (ConstructorInfo)GetMethod());
+			ilgen.Emit(OpCodes.Newobj, stub);
 		}
 #endif // !STUB_GENERATOR
 	}
@@ -1208,6 +1241,19 @@ namespace IKVM.Internal
 			this.fieldType = fieldType;
 			this.field = field;
 			UpdateNonPublicTypeInSignatureFlag();
+#if STATIC_COMPILER
+			if (IsFinal
+				&& DeclaringType.IsPublic
+				&& !DeclaringType.IsInterface
+				&& (IsPublic || (IsProtected && !DeclaringType.IsFinal))
+				&& !DeclaringType.GetClassLoader().StrictFinalFieldSemantics
+				&& DeclaringType is DynamicTypeWrapper
+				&& !(this is ConstantFieldWrapper)
+				&& !(this is DynamicPropertyFieldWrapper))
+			{
+				SetType2FinalField();
+			}
+#endif
 		}
 
 		internal FieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, string name, string sig, ExModifiers modifiers, FieldInfo field)
@@ -1273,7 +1319,19 @@ namespace IKVM.Internal
 			return null;
 #else
 			java.lang.reflect.Field f = (java.lang.reflect.Field)field;
-			return TypeWrapper.FromClass(f.getDeclaringClass()).GetFields()[f._slot()];
+			int slot = f._slot();
+			if (slot == -1)
+			{
+				// it's a Field created by Unsafe.objectFieldOffset(Class,String) so we must resolve based on the name
+				foreach (FieldWrapper fw in TypeWrapper.FromClass(f.getDeclaringClass()).GetFields())
+				{
+					if (fw.Name == f.getName())
+					{
+						return fw;
+					}
+				}
+			}
+			return TypeWrapper.FromClass(f.getDeclaringClass()).GetFields()[slot];
 #endif
 		}
 
@@ -1285,11 +1343,14 @@ namespace IKVM.Internal
 			java.lang.reflect.Field field = reflectionField;
 			if (field == null)
 			{
+				const Modifiers ReflectionFieldModifiersMask = Modifiers.Public | Modifiers.Private | Modifiers.Protected | Modifiers.Static
+					| Modifiers.Final | Modifiers.Volatile | Modifiers.Transient | Modifiers.Synthetic | Modifiers.Enum;
+				Link();
 				field = new java.lang.reflect.Field(
 					this.DeclaringType.ClassObject,
 					this.Name,
-					this.FieldTypeWrapper.ClassObject,
-					(int)this.Modifiers | (this.IsInternal ? 0x40000000 : 0),
+					this.FieldTypeWrapper.EnsureLoadable(this.DeclaringType.GetClassLoader()).ClassObject,
+					(int)(this.Modifiers & ReflectionFieldModifiersMask) | (this.IsInternal ? 0x40000000 : 0),
 					Array.IndexOf(this.DeclaringType.GetFields(), this),
 					this.DeclaringType.GetGenericFieldSignature(this),
 					null
@@ -1541,78 +1602,6 @@ namespace IKVM.Internal
 #endif // !STUB_GENERATOR
 	}
 
-	sealed class GetterFieldWrapper : FieldWrapper
-	{
-		private MethodInfo getter;
-		private PropertyInfo prop;
-
-		// NOTE fi may be null!
-		internal GetterFieldWrapper(TypeWrapper declaringType, TypeWrapper fieldType, FieldInfo fi, string name, string sig, ExModifiers modifiers, MethodInfo getter, PropertyInfo prop)
-			: base(declaringType, fieldType, name, sig, modifiers, fi)
-		{
-			Debug.Assert(!IsVolatile);
-
-			this.getter = getter;
-			this.prop = prop;
-		}
-
-		internal void SetGetter(MethodInfo getter)
-		{
-			this.getter = getter;
-		}
-
-		internal MethodInfo GetGetter()
-		{
-			return getter;
-		}
-
-		internal PropertyInfo GetProperty()
-		{
-			return prop;
-		}
-
-#if !STUB_GENERATOR
-		protected override void EmitGetImpl(CodeEmitter ilgen)
-		{
-			if(!IsStatic && DeclaringType.IsNonPrimitiveValueType)
-			{
-				ilgen.Emit(OpCodes.Unbox, DeclaringType.TypeAsTBD);
-				ilgen.Emit(OpCodes.Call, getter);
-			}
-			else
-			{
-				// NOTE we look at the static-ness of the getter method and not our own,
-				// because for instance fields we can still have a static getter method
-				ilgen.Emit(getter.IsStatic ? OpCodes.Call : OpCodes.Callvirt, getter);
-			}
-		}
-
-		protected override void EmitSetImpl(CodeEmitter ilgen)
-		{
-			if(!IsStatic && DeclaringType.IsNonPrimitiveValueType)
-			{
-				CodeEmitterLocal temp = ilgen.DeclareLocal(FieldTypeWrapper.TypeAsSignatureType);
-				ilgen.Emit(OpCodes.Stloc, temp);
-				ilgen.Emit(OpCodes.Unbox, DeclaringType.TypeAsTBD);
-				ilgen.Emit(OpCodes.Ldloc, temp);
-			}
-			FieldInfo fi = GetField();
-			if(fi != null)
-			{
-				// common case (we're in a DynamicTypeWrapper and the caller is too)
-				ilgen.Emit(IsStatic ? OpCodes.Stsfld : OpCodes.Stfld, fi);
-			}
-			else
-			{
-				// this means that we are an instance on a CompiledTypeWrapper and we're being called
-				// from DynamicMethod based reflection, so we can safely emit a call to the private
-				// setter, because the DynamicMethod is allowed to access our private members.
-				ilgen.Emit(OpCodes.Call, prop.GetSetMethod(true));
-			}
-		}
-#endif // !STUB_GENERATOR
-	}
-
 #if !STUB_GENERATOR
 	// this class represents a .NET property defined in Java with the ikvm.lang.Property annotation
 	sealed class DynamicPropertyFieldWrapper : FieldWrapper
@@ -1708,7 +1697,7 @@ namespace IKVM.Internal
 			// HACK the branch around the throw is to keep the verifier happy
 			CodeEmitterLabel label = ilgen.DefineLabel();
 			ilgen.Emit(OpCodes.Ldc_I4_0);
-			ilgen.Emit(OpCodes.Brtrue_S, label);
+			ilgen.EmitBrtrue(label);
 			ilgen.EmitThrow("java.lang.NoSuchMethodError");
 			ilgen.MarkLabel(label);
 			if (!isStatic)
@@ -1750,7 +1739,7 @@ namespace IKVM.Internal
 			// HACK the branch around the throw is to keep the verifier happy
 			CodeEmitterLabel label = ilgen.DefineLabel();
 			ilgen.Emit(OpCodes.Ldc_I4_0);
-			ilgen.Emit(OpCodes.Brtrue_S, label);
+			ilgen.EmitBrtrue(label);
 			ilgen.EmitThrow("java.lang.NoSuchMethodError");
 			ilgen.MarkLabel(label);
 			ilgen.Emit(OpCodes.Pop);
@@ -1862,7 +1851,7 @@ namespace IKVM.Internal
 				constant is char ||
 				constant is bool)
 			{
-				ilgen.Emit(OpCodes.Ldc_I4, ((IConvertible)constant).ToInt32(null));
+				ilgen.EmitLdc_I4(((IConvertible)constant).ToInt32(null));
 			}
 			else if(constant is string)
 			{
@@ -1870,23 +1859,23 @@ namespace IKVM.Internal
 			}
 			else if(constant is float)
 			{
-				ilgen.Emit(OpCodes.Ldc_R4, (float)constant);
+				ilgen.EmitLdc_R4((float)constant);
 			}
 			else if(constant is double)
 			{
-				ilgen.Emit(OpCodes.Ldc_R8, (double)constant);
+				ilgen.EmitLdc_R8((double)constant);
 			}
 			else if(constant is long)
 			{
-				ilgen.Emit(OpCodes.Ldc_I8, (long)constant);
+				ilgen.EmitLdc_I8((long)constant);
 			}
 			else if(constant is uint)
 			{
-				ilgen.Emit(OpCodes.Ldc_I4, unchecked((int)((IConvertible)constant).ToUInt32(null)));
+				ilgen.EmitLdc_I4(unchecked((int)((IConvertible)constant).ToUInt32(null)));
 			}
 			else if(constant is ulong)
 			{
-				ilgen.Emit(OpCodes.Ldc_I8, unchecked((long)(ulong)constant));
+				ilgen.EmitLdc_I8(unchecked((long)(ulong)constant));
 			}
 			else
 			{
@@ -1934,8 +1923,20 @@ namespace IKVM.Internal
 			return modifiers;
 		}
 
-		private CompiledAccessStubFieldWrapper(TypeWrapper wrapper, PropertyInfo property, TypeWrapper propertyType, string name, string signature, Modifiers modifiers, MemberFlags flags)
-			: base(wrapper, propertyType, name, signature, modifiers, null, flags)
+		// constructor for type 1 access stubs
+		internal CompiledAccessStubFieldWrapper(TypeWrapper wrapper, PropertyInfo property, TypeWrapper propertyType)
+			: this(wrapper, property, null, propertyType, GetModifiers(property), MemberFlags.HideFromReflection | MemberFlags.AccessStub)
+		{
+		}
+
+		// constructor for type 2 access stubs
+		internal CompiledAccessStubFieldWrapper(TypeWrapper wrapper, PropertyInfo property, FieldInfo field, TypeWrapper propertyType)
+			: this(wrapper, property, field, propertyType, AttributeHelper.GetModifiersAttribute(property).Modifiers, MemberFlags.AccessStub)
+		{
+		}
+
+		private CompiledAccessStubFieldWrapper(TypeWrapper wrapper, PropertyInfo property, FieldInfo field, TypeWrapper propertyType, Modifiers modifiers, MemberFlags flags)
+			: base(wrapper, propertyType, property.Name, propertyType.SigName, modifiers, field, flags)
 		{
 			this.getter = property.GetGetMethod(true);
 			this.setter = property.GetSetMethod(true);
@@ -1952,55 +1953,5 @@ namespace IKVM.Internal
 			ilgen.Emit(OpCodes.Call, setter);
 		}
 #endif // !STUB_GENERATOR
-
-		internal static bool TryGet(TypeWrapper wrapper, PropertyInfo property, out FieldWrapper accessStub)
-		{
-			NameSigAttribute nameSig = AttributeHelper.GetNameSig(property);
-			bool hideFromReflection = AttributeHelper.IsHideFromReflection(property);
-
-			if (nameSig != null || hideFromReflection)
-			{
-				TypeWrapper type;
-				string name;
-				string sig;
-				if (nameSig == null)
-				{
-					type = ClassLoaderWrapper.GetWrapperFromType(property.PropertyType);
-					name = property.Name;
-					sig = type.SigName;
-				}
-				else
-				{
-					type = wrapper.GetClassLoader().FieldTypeWrapperFromSigNoThrow(nameSig.Sig);
-					name = nameSig.Name;
-					sig = nameSig.Sig;
-				}
-				Modifiers modifiers;
-				MemberFlags flags = MemberFlags.AccessStub;
-				if (hideFromReflection)
-				{
-					// it's a Type 1 access stub (to make inherited fields visible)
-					flags |= MemberFlags.HideFromReflection;
-					modifiers = GetModifiers(property);
-				}
-				else
-				{
-					// it's a Type 2 access stub (to make fields that have a non-public field type visible)
-					ModifiersAttribute attr = AttributeHelper.GetModifiersAttribute(property);
-					modifiers = attr.Modifiers;
-					if (attr.IsInternal)
-					{
-						flags |= MemberFlags.InternalAccess;
-					}
-				}
-				accessStub = new CompiledAccessStubFieldWrapper(wrapper, property, type, name, sig, modifiers, flags);
-				return true;
-			}
-			else
-			{
-				accessStub = null;
-				return false;
-			}
-		}
 	}
 }
