@@ -1,5 +1,5 @@
 ﻿/*
-  Copyright (C) 2008-2011 Jeroen Frijters
+  Copyright (C) 2008-2012 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -147,5 +147,67 @@ namespace IKVM.Internal
 				return true;
 			}
 		}
+
+		internal static MethodBuilder DefineTypeInitializer(TypeBuilder typeBuilder)
+		{
+			MethodAttributes attr = MethodAttributes.Static | MethodAttributes.RTSpecialName | MethodAttributes.SpecialName;
+			if (typeBuilder.IsInterface)
+			{
+				// LAMESPEC the ECMA spec says (part. I, sect. 8.5.3.2) that all interface members must be public, so we make
+				// the class constructor public.
+				// NOTE it turns out that on .NET 2.0 this isn't necessary anymore (neither Ref.Emit nor the CLR verifier complain about it),
+				// but the C# compiler still considers interfaces with non-public methods to be invalid, so to keep interop with C# we have
+				// to keep making the .cctor method public.
+				attr |= MethodAttributes.Public;
+			}
+			else
+			{
+				attr |= MethodAttributes.Private;
+			}
+			return typeBuilder.DefineMethod(ConstructorInfo.TypeConstructorName, attr, null, Type.EmptyTypes);
+		}
+
+		internal static bool MatchNameAndPublicKeyToken(AssemblyName name1, AssemblyName name2)
+		{
+			return name1.Name.Equals(name2.Name, StringComparison.InvariantCultureIgnoreCase)
+				&& CompareKeys(name1.GetPublicKeyToken(), name2.GetPublicKeyToken());
+		}
+
+		private static bool CompareKeys(byte[] b1, byte[] b2)
+		{
+			int len1 = b1 == null ? 0 : b1.Length;
+			int len2 = b2 == null ? 0 : b2.Length;
+			if (len1 != len2)
+			{
+				return false;
+			}
+			for (int i = 0; i < len1; i++)
+			{
+				if (b1[i] != b2[i])
+				{
+					return false;
+				}
+			}
+			return true;
+		}
+
+		internal static bool IsConstructor(MethodBase method)
+		{
+			return method.IsSpecialName && method.Name == ConstructorInfo.ConstructorName;
+		}
+
+		internal static MethodBuilder DefineConstructor(TypeBuilder tb, MethodAttributes attribs, Type[] parameterTypes)
+		{
+			return tb.DefineMethod(ConstructorInfo.ConstructorName, attribs | MethodAttributes.SpecialName | MethodAttributes.RTSpecialName, null, parameterTypes);
+		}
+
+		internal static bool CanOwnDynamicMethod(Type type)
+		{
+			return type != null
+				&& !type.IsInterface
+				&& !type.HasElementType
+				&& !type.IsGenericTypeDefinition
+				&& !type.IsGenericParameter;
+		}
 	}
 }
