@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 2002, 2003, 2004, 2005, 2006 Jeroen Frijters
+  Copyright (C) 2002-2011 Jeroen Frijters
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -67,7 +67,7 @@ namespace IKVM.Attributes
 			this.table = table;
 		}
 
-		public class LineNumberWriter
+		public sealed class LineNumberWriter
 		{
 			private System.IO.MemoryStream stream;
 			private int prevILOffset;
@@ -420,7 +420,7 @@ namespace IKVM.Attributes
 	{
 	}
 
-	[AttributeUsage(AttributeTargets.Method | AttributeTargets.Property)]
+	[AttributeUsage(AttributeTargets.Method | AttributeTargets.Field | AttributeTargets.Property)]
 	public sealed class HideFromReflectionAttribute : Attribute
 	{
 	}
@@ -786,4 +786,7 @@ namespace IKVM.Attributes
 			return packages;
 		}
 	}
+
+	// used in custom modifier for access stubs
+	public static class AccessStub { }
 }
