@@ -140,6 +140,11 @@ namespace IKVM.NativeCode.ikvm.@internal
 	{
 		static class StubGenerator
 		{
+			public static int getRealModifiers(jlClass c)
+			{
+				return (int)TypeWrapper.FromClass(c).Modifiers;
+			}
+
 			public static string getAssemblyName(jlClass c)
 			{
 				TypeWrapper wrapper = TypeWrapper.FromClass(c);
@@ -168,11 +173,6 @@ namespace IKVM.NativeCode.ikvm.@internal
 				{
 					return fi.IsDefined(typeof(ObsoleteAttribute), false);
 				}
-				GetterFieldWrapper getter = fieldWrapper as GetterFieldWrapper;
-				if(getter != null)
-				{
-					return getter.GetProperty().IsDefined(typeof(ObsoleteAttribute), false);
-				}
 				return false;
 			}
 
@@ -239,6 +239,20 @@ namespace IKVM.NativeCode.ikvm.runtime
 			}
 		}
 
+		public static global::java.net.URL GetManifest(Assembly assembly)
+		{
+#if FIRST_PASS
+			return null;
+#else
+			IKVM.Internal.AssemblyClassLoader wrapper = IKVM.Internal.AssemblyClassLoader.FromAssembly(assembly);
+			foreach (global::java.net.URL url in wrapper.FindResources("META-INF/MANIFEST.MF"))
+			{
+				return url;
+			}
+			return null;
+#endif
+		}
+
 		public static global::java.net.URL getResource(global::java.lang.ClassLoader classLoader, Assembly assembly, string name)
 		{
 #if FIRST_PASS
@@ -490,6 +504,25 @@ namespace IKVM.NativeCode.ikvm.runtime
 			return null;
 		}
 
+		public static jlClass getClassFromTypeHandle(RuntimeTypeHandle handle, int rank)
+		{
+			Type t = Type.GetTypeFromHandle(handle);
+			if(t.IsPrimitive || ClassLoaderWrapper.IsRemappedType(t) || t == typeof(void))
+			{
+				return DotNetTypeWrapper.GetWrapperFromDotNetType(t).MakeArrayType(rank).ClassObject;
+			}
+			if(!IsVisibleAsClass(t))
+			{
+				return null;
+			}
+			TypeWrapper tw = ClassLoaderWrapper.GetWrapperFromType(t);
+			if(tw != null)
+			{
+				return tw.MakeArrayType(rank).ClassObject;
+			}
+			return null;
+		}
+
 		public static jlClass getFriendlyClassFromType(Type type)
 		{
 			int rank = 0;
